# Builtins
import re
from getpass import getpass
import os

# Source: https://stackoverflow.com/questions/287871/how-to-print-colored-text-to-the-terminal
HEADER = "\033[95m"
OKBLUE = "\033[94m"
OKCYAN = "\033[96m"
OKGREEN = "\033[92m"
WARNING = "\033[93m"
FAIL = "\033[91m"
ENDC = "\033[0m"
BOLD = "\033[1m"
UNDERLINE = "\033[4m"


class Wizard:
    """
    The base class for all wizards that holds functions to get user inputs.
    Designed so that the *only* function the child classes should need to
    implement is the `create_secret` function.
    """

    def __init__(self):
        self.update_size()

        self.text_counter = 0
        self.prompt_counter = 0

    def update_size(self):
        # https://stackoverflow.com/questions/566746/how-to-get-linux-console-window-width-in-python
        with os.popen("stty size", "r") as dim:
            # with closes the subprocess opened by popen
            dimensions = dim.read().split()
        # If dimensions not given, i.e. not run in a terminal such as on github actions
        self.rows, self.columns = dimensions if len(dimensions) == 2 else (0, 0)
        self.columns = int(self.columns)
        self.rows = int(self.rows)

    def reset_counter(self) -> None:
        self.text_counter = 0
        self.prompt_counter = 0

    def print(self, text: str) -> None:
        if self.text_counter == 0:
            self.reset_counter()
            header = "💬 "
        else:
            header = "  "
        print(f"{header}{text}", end="")
        self.text_counter += 1

    def println(self, text: str) -> None:
        if self.text_counter == 0:
            self.reset_counter()
            header = "💬 "
        else:
            header = "  "
        print(f"{header}{text}")
        self.text_counter += 1

    def get_bool(
        self,
        prompt: str = "y/n",
        true_pat: str = r"y|yes",
        false_pat: str = r"n|no",
        default: str = None,
        persistent=True,
    ) -> bool:
        """
        Prompts the user for a binary decision. Ignores case in regex matching. Return False if the input does not match any pattern and persistent is False.

        :prompt: What to ask the user
        :true_pat: A (regex) pattern that is used to validate the user input is True
        :false_pat: A (regex) pattern that is used to validate the user input is False
        :default: A string that should return True when matched with its associated value's pattern. If None then ignored.
        :persistent: A bool that if True, and if default is None, will continue to prompt the user.
        """
        self.reset_counter()

        df = ""
        if default is not None:
            df = f"[{default}]"

        prompt = f"❓ {HEADER}{prompt} (y/n){df}{ENDC} "

        value = input(prompt)

        if re.fullmatch(true_pat, value, flags=re.IGNORECASE):
            return True

        elif re.fullmatch(false_pat, value, flags=re.IGNORECASE):
            return False

        if default is not None:
            return re.fullmatch(true_pat, default, flags=re.IGNORECASE)

        else:
            if persistent:
                return self.get_bool(prompt, true_pat, false_pat, default, persistent)
            else:
                return False

    def get_string(
        self, prompt: str = "input", pattern: str = ".+", persistent: bool = False
    ) -> str:
        """
        Prompts the user for any string, with an optional pattern, ignoring case. If persistent is True then will continuly prompt the user for input.

        :prompt: What to ask the user
        :pattern: A (regex) pattern that is used to validate the user input. Default to any string with at least 1 input
        :persistent:A bool that if True, will continue to prompt the user.
        """
        if self.prompt_counter == 0:
            self.reset_counter()
            print("❓ ", end="")

        value = input(f"{HEADER}{prompt}{ENDC}\n\t➔")

        if pattern is None or not persistent:
            return value
        elif re.fullmatch(pattern, value, flags=re.IGNORECASE):
            return value
        else:
            return self.get_string(prompt, pattern, persistent)

    def get_int(
        self, prompt="input number", default: int = None, persistent: bool = False
    ) -> int:
        """
        Prompts the user for an integer.

        :prompt: What to ask the user
        :default: If persistent is off, then this retuns if the user fails to specify a number.
        :persistent: If True then keep prompting the user for a valid input.
        """
        value = self.get_string(prompt, r"\d+", persistent)

        try:
            return int(value)
        except ValueError as e:
            print(f"Invalid input {value}, using {default} instead!\nError: {e}")
            return default

    def get_password(self, prompt="Password: ") -> str:
        """
        Prompts the user for a password.
        """
        return getpass(prompt)

    def wait_for_input(self) -> None:
        """
        Prompts the user for any input and to give the user the ability to do something while the program waits.
        """
        print("Press Enter...".rjust(self.columns))
        input("")



# Builtins
import yaml
import threading
import datetime as dt
from typing import Any, Dict, List, Tuple

# External libraries
import pandas as pd
from alpaca_trade_api.rest import REST, TimeFrame, URL
from alpaca_trade_api import Stream

# Submodule imports
from harvest.api._base import StreamAPI, API
from harvest.utils import *


class Alpaca(StreamAPI):

    interval_list = [
        Interval.MIN_1,
    ]

    def __init__(
        self,
        path: str = None,
        is_basic_account: bool = False,
        paper_trader: bool = False,
    ):
        super().__init__(path)
        self.basic = is_basic_account

        endpoint = (
            "https://paper-api.alpaca.markets"
            if paper_trader
            else "https://api.alpaca.markets"
        )
        self.api = REST(self.config["api_key"], self.config["secret_key"], endpoint)

        data_feed = "iex" if self.basic else "sip"
        self.stream = Stream(
            self.config["api_key"],
            self.config["secret_key"],
            URL(endpoint),
            data_feed=data_feed,
        )
        self.data_lock = threading.Lock()
        self.data = {}

    async def update_data(self, bar):
        # Update data with the latest bars
        self.data_lock.acquire()
        bar = bar.__dict__["_raw"]
        symbol = bar["symbol"]
        df = pd.DataFrame(
            [
                {
                    "t": bar["timestamp"],
                    "o": bar["open"],
                    "h": bar["high"],
                    "l": bar["low"],
                    "c": bar["close"],
                    "v": bar["volume"],
                }
            ]
        )

        symbol = f"@{symbol}" if is_crypto(symbol) else symbol
        debugger.info(f"Got data for {symbol}")
        df = self._format_df(df, symbol)
        self.data[symbol] = df
        if set(self.data.keys()) == set(self.watch_stock + self.watch_crypto):
            data = self.data
            self.data = {}
            self.data_lock.release()
            self.trader_main(data)
        else:
            self.data_lock.release()

    def setup(self, interval: Dict, trader_main=None):
        super().setup(interval, trader_main)

        self.watch_stock = []
        self.watch_crypto = []
        cryptos = []

        for s in interval:
            if is_crypto(s):
                self.watch_crypto.append(s)
                cryptos.append(s[1:])
            else:
                self.watch_stock.append(s)

        self.stream.on_bar(*(self.watch_stock + cryptos))(self.update_data)

        self.option_cache = {}

    def start(self):
        self.stream.run()

    def exit(self):
        self.option_cache = {}

    # -------------- Streamer methods -------------- #

    @API._exception_handler
    def fetch_price_history(
        self,
        symbol: str,
        interval: Interval,
        start: dt.datetime = None,
        end: dt.datetime = None,
    ):

        debugger.debug(f"Fetching {symbol} {interval} price history")

        if start is None:
            start = now() - dt.timedelta(days=365 * 5)
        if end is None:
            end = now()

        if start >= end:
            return pd.DataFrame()

        df = self.get_data_from_alpaca(symbol, interval, start, end)
        return df

    @API._exception_handler
    def fetch_chain_info(self, symbol: str):
        raise NotImplementedError("Alpaca does not support options.")

    @API._exception_handler
    def fetch_chain_data(self, symbol: str, date: dt.datetime):
        raise NotImplementedError("Alpaca does not support options.")

    @API._exception_handler
    def fetch_option_market_data(self, occ_symbol: str):
        raise NotImplementedError("Alpaca does not support options.")

    # ------------- Broker methods ------------- #

    @API._exception_handler
    def fetch_stock_positions(self):
        def fmt(stock: Dict[str, Any]):
            return {
                "symbol": stock["symbol"],
                "avg_price": float(stock["avg_entry_price"]),
                "quantity": float(stock["qty"]),
                "alpaca": stock,
            }

        return [
            fmt(pos.__dict__["_raw"])
            for pos in self.api.list_positions()
            if pos.asset_class != "crypto"
        ]

    @API._exception_handler
    def fetch_option_positions(self):
        debugger.error("Alpaca does not support options. Returning an empty list.")
        return []

    @API._exception_handler
    def fetch_crypto_positions(self, key=None):
        if self.basic:
            debugger.error(
                "Alpaca basic accounts do not support crypto. Returning an empty list."
            )
            return []

        def fmt(crypto: Dict[str, Any]):
            return {
                "symbol": "@" + crypto["symbol"],
                "avg_price": float(crypto["avg_entry_price"]),
                "quantity": float(stock["qty"]),
                "alpaca": crypto,
            }

        return [
            fmt(pos.__dict__["_raw"])
            for pos in self.api.list_positions()
            if pos.asset_class == "crypto"
        ]

    @API._exception_handler
    def update_option_positions(self, positions: List[Any]):
        debugger.error("Alpaca does not support options. Doing nothing.")

    @API._exception_handler
    def fetch_account(self):
        account = self.api.get_account().__dict__["_raw"]
        return {
            "equity": float(account["equity"]),
            "cash": float(account["cash"]),
            "buying_power": float(account["buying_power"]),
            "multiplier": float(account["multiplier"]),
            "alpaca": account,
        }

    @API._exception_handler
    def fetch_stock_order_status(self, id: str):
        return self.api.get_order(id).__dict__["_raw"]

    @API._exception_handler
    def fetch_option_order_status(self, id):
        raise NotImplementedError("Alpaca does not support options.")

    @API._exception_handler
    def fetch_crypto_order_status(self, id):
        if self.basic:
            raise Exception("Alpaca basic accounts do not support crypto.")
        return self.api.get_order(id).__dict__["_raw"]

    @API._exception_handler
    def fetch_order_queue(self):
        return [
            self.format_order_status(pos.__dict__["_raw"])
            for pos in self.api.list_orders()
        ]

    # --------------- Methods for Trading --------------- #

    def order_stock_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):

        return self.api.submit_order(
            symbol,
            quantity,
            side=side,
            type="limit",
            limit_price=limit_price,
            time_in_force=in_force,
            extended_hours=extended,
        )

    def order_crypto_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        if self.basic:
            raise Exception("Alpaca basic accounts do not support crypto.")

        symbol = symbol[1:]

        order = self.api.submit_order(
            asset,
            quantity,
            side=side,
            type="limit",
            limit_price=limit_price,
            time_in_force=in_force,
            extended_hours=extended,
        ).__dict__["_raw"]

        return {
            "type": "CRYPTO" if is_crypto(symbol) else "STOCK",
            "id": order["id"],
            "symbol": symbol,
            "alpaca": order,
        }

    def order_option_limit(
        self,
        side: str,
        symbol: str,
        quantity: int,
        limit_price: float,
        option_type,
        exp_date: dt.datetime,
        strike,
        in_force: str = "gtc",
    ):
        raise NotImplementedError("Alpaca does not support options.")

    # ------------- Helper methods ------------- #

    def format_order_status(self, order: Dict[str, Any], is_stock: bool = True):
        return {
            "type": "STOCK" if is_stock else "CRYPTO",
            "id": order["id"],
            "symbol": ("" if is_stock else "@") + order["symbol"],
            "quantity": float(order["qty"]),
            "filled_quantity": float(order["filed_qty"]),
            "side": order["side"],
            "time_in_force": order["time_in_force"],
            "status": order["status"],
            "alpaca": order,
        }

    def get_data_from_alpaca(
        self,
        symbol: str,
        interval: Interval,
        start: dt.datetime,
        end: dt.datetime,
    ) -> pd.DataFrame:
        if self.basic and is_crypto(symbol):
            debugger.error(
                "Alpaca basic accounts do not support crypto. Returning empty dataframe"
            )
            return pd.DataFrame()

        current_time = now()
        if self.basic and start < current_time - dt.timedelta(days=365 * 5):
            debugger.warning(
                "Start time is over five years old! Only data from the past five years will be returned for basic accounts."
            )
            start = current_time - dt.timedelta(days=365 * 5)

        if self.basic and end >= current_time - dt.timedelta(minutes=15):
            debugger.warning(
                "End time is less than 15 minutes old! Only data over 15 minutes old will be returned for basic accounts."
            )
            end = current_time - dt.timedelta(minutes=15)

        timespan = expand_interval(interval)[1]
        if timespan == "MIN":
            timespan = "1Min"
        elif timespan == "HR":
            timespan = "1Hour"
        elif timespan == "DAY":
            timespan = "1Day"

        start_str = start.isoformat()
        end_str = end.isoformat()

        temp_symbol = symbol[1:] if is_crypto(symbol) else symbol
        bars = self.api.get_bars(
            temp_symbol, TimeFrame(timespan), start_str, end_str, adjustment="raw"
        )
        df = pd.DataFrame((bar.__dict__["_raw"] for bar in bars))
        df = self._format_df(df, symbol)
        df = aggregate_df(df, interval)
        return df

    def _format_df(self, df: pd.DataFrame, symbol: str):
        df = df[["t", "o", "h", "l", "c", "v"]]
        df.rename(
            columns={
                "t": "timestamp",
                "o": "open",
                "h": "high",
                "l": "low",
                "c": "close",
                "v": "volume",
            },
            inplace=True,
        )

        df["timestamp"] = pd.to_datetime(df["timestamp"], utc=True)
        df.set_index("timestamp", inplace=True)

        df.columns = pd.MultiIndex.from_product([[symbol], df.columns])

        return df.dropna()

    def create_secret(self, path: str) -> bool:
        import harvest.wizard as wizard

        w = wizard.Wizard()

        w.println("Hmm, looks like you haven't set up an api key for Alpaca.")
        should_setup = w.get_bool("Do you want to set it up now?", default="y")

        if not should_setup:
            w.println("You can't use Alpaca without an API key.")
            w.println(
                "You can set up the credentials manually, or use other streamers."
            )
            return False

        w.println("Alright! Let's get started")

        have_account = w.get_bool("Do you have an Alpaca account?", default="y")
        if not have_account:
            w.println(
                "In that case you'll first need to make an account. This takes a few steps."
            )
            w.println(
                "First visit: https://alpaca.markets/ and sign up. Hit Enter or Return for the next step."
            )
            w.wait_for_input()
            w.println("Follow the setups to make an individual or buisness account.")
            w.wait_for_input()
            w.println(
                "In the sidebar of the Alpaca Dashboard, note if you want to use a live account (real money) or a paper account (simulated)."
            )
            w.wait_for_input()
            w.println(
                "On the right-hand side, in the Your API Keys box, click View and then Generate API Key. Copy these somewhere safe."
            )
            w.wait_for_input()

        api_key_id = w.get_string("Enter your API key ID")
        secret_key = w.get_password("Enter your API secret key")

        w.println(f"All steps are complete now 🎉. Generating {path}...")

        d = {"api_key": f"{api_key_id}", "secret_key": f"{secret_key}"}

        with open(path, "w") as file:
            yml = yaml.dump(d, file)

        w.println(
            f"{path} has been created! Make sure you keep this file somewhere secure and never share it with other people."
        )

        return True



import os
import sys
import inspect
import argparse
import importlib.util

from os import listdir
from os.path import isfile, join

# Lambda functions cannot raise exceptions so using higher order functions.
def _raise(e):
    def raise_helper():
        raise e

    return raise_helper


from harvest.storage.base_storage import BaseStorage
from harvest.storage.csv_storage import CSVStorage
from harvest.storage.pickle_storage import PickleStorage

# For imports that have extra dependencies, surpress the import error unless the user specifics that resource and does not have the depenedencies.
try:
    from harvest.storage.database_storage import DBStorage
except ModuleNotFoundError as e:
    DBStorage = _raise(e)

from harvest.api.dummy import DummyStreamer
from harvest.api.yahoo import YahooStreamer
from harvest.api.polygon import PolygonStreamer
from harvest.api.paper import PaperBroker

try:
    from harvest.api.robinhood import Robinhood
except ModuleNotFoundError as e:
    Robinhood = _raise(e)
try:
    from harvest.api.alpaca import Alpaca
except ModuleNotFoundError as e:
    Alpaca = _raise(e)
try:
    from harvest.api.kraken import Kraken
except ModuleNotFoundError as e:
    Kraken = _raise(e)
try:
    from harvest.api.webull import Webull
except ModuleNotFoundError as e:
    Webull = _raise(e)

from harvest.trader import LiveTrader
from harvest.algo import BaseAlgo

storages = {
    "memory": BaseStorage,
    "csv": CSVStorage,
    "pickle": PickleStorage,
    "db": DBStorage,
}

streamers = {
    "dummy": DummyStreamer,
    "yahoo": YahooStreamer,
    "polygon": PolygonStreamer,
    "robinhood": Robinhood,
    "alpaca": Alpaca,
    "kraken": Kraken,
    "webull": Webull,
}

brokers = {
    "paper": PaperBroker,
    "robinhood": Robinhood,
    "alpaca": Alpaca,
    "kraken": Kraken,
    "webull": Webull,
}

parser = argparse.ArgumentParser(description="Harvest CLI")
subparsers = parser.add_subparsers(dest="command")

# Parser for starting harvest
start_parser = subparsers.add_parser("start")
start_parser.add_argument(
    "-o",
    "--storage",
    default="memory",
    help="the way to store asset data",
    choices=list(storages.keys()),
)
start_parser.add_argument(
    "-s",
    "--streamer",
    default="yahoo",
    help="fetches asset data",
    choices=list(streamers.keys()),
)
start_parser.add_argument(
    "-b",
    "--broker",
    default="paper",
    help="buys and sells assets on your behalf",
    choices=list(brokers.keys()),
)

# Directory with algos that you want to run, default is the current working directory.
start_parser.add_argument(
    "-d",
    "--directory",
    default=".",
    help="directory where algorithms are located",
)
start_parser.add_argument(
    "--debug", default=False, action=argparse.BooleanOptionalAction
)


# Parser for visualing data
visualize_parser = subparsers.add_parser("visualize")
visualize_parser.add_argument("path", help="path to harvest generated data file")


def main():
    """
    Entrypoint which parses the command line arguments. Calls subcommands based on which subparser was used.
    :args: A Namespace object containing parsed user arguments.
    """
    args = parser.parse_args()

    # Handles the start command
    if args.command == "start":
        start(args)
    elif args.command == "visualize":
        visualize(args)
    # Show help if case not found
    else:
        parser.print_help(sys.stderr)
        sys.exit(1)


def start(args: argparse.Namespace, test: bool = False):
    """
    Starts the Harvest LiveTrader with the given storage, streamer, broker, and algos specified.
    :args: A Namespace object containing parsed user arguments.
    :test: True if we are testing so that we can exit this function cleanly.
    """
    storage = _get_storage(args.storage)
    streamer = _get_streamer(args.streamer)
    broker = _get_broker(args.broker)
    debug = args.debug
    trader = LiveTrader(streamer=streamer, broker=broker, storage=storage, debug=debug)

    # Get the directories.
    directory = args.directory
    print(f"Searching directory {directory}")
    files = [fi for fi in listdir(directory) if isfile(join(directory, fi))]
    print(f"Found files {files}")
    # For each file in the directory...
    for f in files:
        names = f.split(".")
        # Filter out non-python files.
        if len(names) <= 1 or names[-1] != "py":
            continue
        name = "".join(names[:-1])

        # ...open it...
        with open(join(directory, f), "r") as algo_file:
            firstline = algo_file.readline()
            if firstline.find("HARVEST_SKIP") != -1:
                print(f"Skipping {f}")
                continue

        # ...load in the entire file and add the algo to the trader.
        algo_path = os.path.realpath(join(directory, f))
        spec = importlib.util.spec_from_file_location(name, algo_path)
        algo = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(algo)
        # Iterate though the variables and if a variable is a subclass of BaseAlgo instantiate it and added to the trader.
        for algo_cls in inspect.getmembers(algo):
            k, v = algo_cls[0], algo_cls[1]
            if inspect.isclass(v) and v != BaseAlgo and issubclass(v, BaseAlgo):
                print(f"Found algo {k} in {f}, adding to trader")
                trader.add_algo(v())

    if not test:
        trader.start()


def visualize(args: argparse.Namespace):
    """
    Read a csv or pickle file created by Harvest with ohlc data and graph the data.
    :args: A Namespace object containing parsed user arguments.
    """
    import re
    import pandas as pd
    import mplfinance as mpf

    # Open the file using the appropriate parser.
    if args.path.endswith(".csv"):
        df = pd.read_csv(args.path)
        df["timestamp"] = pd.to_datetime(df["timestamp"])
        df.set_index("timestamp", inplace=True)
    elif args.path.endswith(".pickle"):
        df = pd.read_pickle(args.path)
    else:
        print("Invalid file extension, expecting .csv or .pickle.")
        sys.exit(1)

    if df.empty:
        print(f"No data found in {args.path}.")
        sys.exit(2)

    path = os.path.basename(args.path)
    # File names are asset {ticker name}@{interval}.{file format}
    file_search = re.search("^(@?[\w]+)@([\w]+).(csv|pickle)$", path)
    symbol, interval = file_search.group(1), file_search.group(2)
    open_price = df.iloc[0]["open"]
    close_price = df.iloc[-1]["close"]
    high_price = df["high"].max()
    low_price = df["low"].min()
    price_delta = close_price - open_price
    price_delta_precent = 100 % (price_delta / open_price)
    volume = df["volume"].sum()

    print(f"{symbol} at {interval}")
    print(f"open\t{open_price}")
    print(f"high\t{high_price}")
    print(f"low\t{low_price}")
    print(f"close\t{close_price}")
    print(f"price change\t{price_delta}")
    print(f"price change percentage\t{price_delta_precent}%")
    print(f"volume\t{volume}")
    mpf.plot(
        df,
        type="candle",
        style="charles",
        volume=True,
        show_nontrading=True,
        title=path,
    )


def _get_storage(storage: str):
    """
    Returns the storage instance specified by the user.
    :storage: The type of storage to be instantiated.
    """
    storage_cls = storages.get(storage)
    if storage_cls is None:
        raise ValueError(
            f"Invalid storage option: {storage}, valid options are {storages.keys()}"
        )
    return storage_cls()


def _get_streamer(streamer):
    """
    Returns the storage instance specified by the user.
    :streamer: The type of streamer to be instantiated.
    """
    streamer_cls = streamers.get(streamer)
    if streamer_cls is None:
        raise ValueError(
            f"Invalid streamer option: {streamer}, valid options are {streamers.keys()}"
        )
    return streamer_cls()


def _get_broker(broker):
    """
    Returns the storage instance specified by the user.
    :broker: The type of broker to be instantiated.
    """
    broker_cls = brokers.get(broker)
    if broker_cls is None:
        raise ValueError(
            f"Invalid broker option: {broker}, valid options are: {brokers.keys()}"
        )
    return broker_cls()


if __name__ == "__main__":
    main()



# Builtins
import unittest

import pandas as pd
from pandas.testing import assert_frame_equal

from harvest.storage import BaseStorage
from harvest.utils import *


class TestBaseStorage(unittest.TestCase):
    def test_create_storage(self):
        storage = BaseStorage()
        self.assertEqual(storage.storage_price, {})

    def test_simple_store(self):
        storage = BaseStorage()
        data = gen_data("A", 50)
        storage.store("A", Interval.MIN_1, data.copy(True))

        self.assertTrue(not pd.isna(data.iloc[0]["A"]["low"]))
        self.assertEqual(list(storage.storage_price.keys()), ["A"])
        self.assertEqual(list(storage.storage_price["A"].keys()), [Interval.MIN_1])

    def test_simple_load(self):
        storage = BaseStorage()
        data = gen_data("A", 50)
        storage.store("A", Interval.MIN_1, data.copy(True))
        loaded_data = storage.load("A", Interval.MIN_1)

        assert_frame_equal(loaded_data, data)

    def test_store_no_overlap(self):
        storage = BaseStorage()
        data = gen_data("A", 100)
        storage.store("A", Interval.MIN_1, data.copy(True).iloc[:50])
        storage.store("A", Interval.MIN_1, data.copy(True).iloc[50:])
        loaded_data = storage.load("A", Interval.MIN_1)

        self.assertTrue(not pd.isnull(data.iloc[0]["A"]["low"]))
        self.assertTrue(not pd.isnull(loaded_data.iloc[0]["A"]["low"]))
        assert_frame_equal(loaded_data, data)

    def test_store_overlap1(self):
        storage = BaseStorage()
        data = gen_data("A", 100)
        storage.store("A", Interval.MIN_1, data.copy(True).iloc[:75])
        storage.store("A", Interval.MIN_1, data.copy(True).iloc[25:])
        loaded_data = storage.load("A", Interval.MIN_1)

        self.assertTrue(not pd.isnull(data.iloc[0]["A"]["low"]))
        self.assertTrue(not pd.isnull(loaded_data.iloc[0]["A"]["low"]))
        assert_frame_equal(loaded_data, data)

    def test_store_overlap2(self):
        storage = BaseStorage()
        data = gen_data("A", 100)
        storage.store("A", Interval.MIN_1, data.copy(True).iloc[25:])
        storage.store("A", Interval.MIN_1, data.copy(True).iloc[:75])
        loaded_data = storage.load("A", Interval.MIN_1)

        self.assertTrue(not pd.isnull(data.iloc[0]["A"]["low"]))
        self.assertTrue(not pd.isnull(loaded_data.iloc[0]["A"]["low"]))
        assert_frame_equal(loaded_data, data)

    def test_store_within(self):
        storage = BaseStorage()
        data = gen_data("A", 100)
        storage.store("A", Interval.MIN_1, data.copy(True).iloc[25:75])
        storage.store("A", Interval.MIN_1, data.copy(True))
        loaded_data = storage.load("A", Interval.MIN_1)

        self.assertTrue(not pd.isnull(data.iloc[0]["A"]["low"]))
        self.assertTrue(not pd.isnull(loaded_data.iloc[0]["A"]["low"]))
        assert_frame_equal(loaded_data, data)

    def test_store_over(self):
        storage = BaseStorage()
        data = gen_data("A", 100)
        storage.store("A", Interval.MIN_1, data.copy(True))
        storage.store("A", Interval.MIN_1, data.copy(True).iloc[25:75])
        loaded_data = storage.load("A", Interval.MIN_1)

        self.assertTrue(not pd.isnull(data.iloc[0]["A"]["low"]))
        self.assertTrue(not pd.isnull(loaded_data.iloc[0]["A"]["low"]))
        assert_frame_equal(loaded_data, data)

    def test_load_no_interval(self):
        storage = BaseStorage()
        data = gen_data("A", 50)
        storage.store("A", Interval.MIN_1, data.copy(True))
        loaded_data = storage.load("A")

        assert_frame_equal(loaded_data, data)

    def test_store_gap(self):
        storage = BaseStorage()
        data = gen_data("A", 100)
        storage.store("A", Interval.MIN_1, data.copy(True).iloc[:25])
        storage.store("A", Interval.MIN_1, data.copy(True).iloc[75:])
        loaded_data = storage.load("A", Interval.MIN_1)

        self.assertTrue(not pd.isnull(data.iloc[0]["A"]["low"]))
        self.assertTrue(not pd.isnull(loaded_data.iloc[0]["A"]["low"]))
        assert_frame_equal(loaded_data, data.iloc[:25].append(data.iloc[75:]))

    # def test_agg_load(self):
    #     storage = BaseStorage()
    #     data = gen_data("A", 100)
    #     storage.store("A", Interval.MIN_1, data.copy(True))
    #     loaded_data = storage.load("A", Interval.MIN_1)

    #     self.assertTrue(not pd.isnull(data.iloc[0]["A"]["low"]))
    #     self.assertTrue(not pd.isnull(loaded_data.iloc[0]["A"]["low"]))
    #     self.assertTrue(loaded_data.shape, (20, 5))


if __name__ == "__main__":
    unittest.main()



import unittest


class TestRHBroker(unittest.TestCase):
    def test_trade(self):
        """
        Test that it can buy and sell crypto
        """
        pass


if __name__ == "__main__":
    unittest.main()



# Builtins
import unittest

from harvest.plugin._base import Plugin


class TestPlugin(unittest.TestCase):
    def test_init(self):
        plugin = Plugin("my_plugin", ["pandas", "finta", "yaml"])
        self.assertEqual(plugin.name, "my_plugin")


if __name__ == "__main__":
    unittest.main()



# Builtins
import unittest
import time

# Submodule imports
from harvest.utils import *

import datetime as dt


class TestUtils(unittest.TestCase):
    def test_is_freq(self):
        """If invalid aggregation is set, it should raise an error"""

        # If interval is '1MIN', it should always return True
        self.assertTrue(is_freq(dt.datetime(2000, 1, 1, 0, 0, 0), Interval.MIN_1))
        self.assertTrue(is_freq(dt.datetime(2000, 1, 1, 1, 1, 1), Interval.MIN_1))

        # If interval is '5MIN', it should return True every 5 minutes
        self.assertTrue(is_freq(dt.datetime(2000, 1, 1, 0, 0, 0), Interval.MIN_5))
        self.assertTrue(is_freq(dt.datetime(2000, 1, 1, 0, 35, 0), Interval.MIN_5))
        self.assertFalse(is_freq(dt.datetime(2000, 1, 1, 1, 59, 0), Interval.MIN_5))

        # If interval is '30MIN', it should return True every 30 minutes
        self.assertTrue(is_freq(dt.datetime(2000, 1, 1, 0, 0, 0), Interval.MIN_30))
        self.assertTrue(is_freq(dt.datetime(2000, 1, 1, 0, 30, 0), Interval.MIN_30))
        self.assertFalse(is_freq(dt.datetime(2000, 1, 1, 1, 35, 0), Interval.MIN_30))

        # If interval is '1HR', it should return True every hour
        self.assertTrue(is_freq(dt.datetime(2000, 1, 1, 0, 0, 0), Interval.HR_1))
        self.assertTrue(is_freq(dt.datetime(2000, 1, 1, 1, 0, 0), Interval.HR_1))
        self.assertFalse(is_freq(dt.datetime(2000, 1, 1, 1, 40, 0), Interval.HR_1))


if __name__ == "__main__":
    unittest.main()



# Builtins
import pathlib
import unittest
import datetime as dt

from harvest.api.paper import PaperBroker
from harvest.api.dummy import DummyStreamer

from harvest.utils import *


class TestPaperBroker(unittest.TestCase):
    def test_account(self):
        dummy = PaperBroker()
        d = dummy.fetch_account()
        self.assertEqual(d["equity"], 1000000.0)
        self.assertEqual(d["cash"], 1000000.0)
        self.assertEqual(d["buying_power"], 1000000.0)
        self.assertEqual(d["multiplier"], 1)

    def test_dummy_account(self):
        directory = pathlib.Path(__file__).parent.resolve()
        dummy = PaperBroker(str(directory) + "/../dummy_account.yaml")
        stocks = dummy.fetch_stock_positions()
        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0]["symbol"], "A")
        self.assertEqual(stocks[0]["avg_price"], 1.0)
        self.assertEqual(stocks[0]["quantity"], 5)

        cryptos = dummy.fetch_crypto_positions()
        self.assertEqual(len(cryptos), 1)
        self.assertEqual(cryptos[0]["symbol"], "@C")
        self.assertEqual(cryptos[0]["avg_price"], 289.21)
        self.assertEqual(cryptos[0]["quantity"], 2)

    def test_buy_order_limit(self):
        dummy = PaperBroker()
        interval = {"A": {"interval": Interval.MIN_1, "aggregations": []}}
        dummy.setup(interval)
        order = dummy.order_stock_limit("buy", "A", 5, 50000)
        self.assertEqual(order["type"], "STOCK")
        self.assertEqual(order["id"], 0)
        self.assertEqual(order["symbol"], "A")

        status = dummy.fetch_stock_order_status(order["id"])
        self.assertEqual(status["id"], 0)
        self.assertEqual(status["symbol"], "A")
        self.assertEqual(status["quantity"], 5)
        self.assertEqual(status["filled_qty"], 5)
        self.assertEqual(status["side"], "buy")
        self.assertEqual(status["time_in_force"], "gtc")
        self.assertEqual(status["status"], "filled")

    def test_buy(self):
        dummy = PaperBroker()
        interval = {"A": {"interval": Interval.MIN_1, "aggregations": []}}
        dummy.setup(interval)
        order = dummy.buy("A", 5, 1e5)
        self.assertEqual(order["type"], "STOCK")
        self.assertEqual(order["id"], 0)
        self.assertEqual(order["symbol"], "A")

        status = dummy.fetch_stock_order_status(order["id"])
        self.assertEqual(status["id"], 0)
        self.assertEqual(status["symbol"], "A")
        self.assertEqual(status["quantity"], 5)
        self.assertEqual(status["filled_qty"], 5)
        self.assertEqual(status["side"], "buy")
        self.assertEqual(status["time_in_force"], "gtc")
        self.assertEqual(status["status"], "filled")

    def test_sell_order_limit(self):
        directory = pathlib.Path(__file__).parent.resolve()
        dummy = PaperBroker(str(directory) + "/../dummy_account.yaml")
        interval = {"A": {"interval": Interval.MIN_1, "aggregations": []}}
        dummy.setup(interval)
        order = dummy.order_stock_limit("sell", "A", 2, 50000)
        self.assertEqual(order["type"], "STOCK")
        self.assertEqual(order["id"], 0)
        self.assertEqual(order["symbol"], "A")

        status = dummy.fetch_stock_order_status(order["id"])
        self.assertEqual(status["id"], 0)
        self.assertEqual(status["symbol"], "A")
        self.assertEqual(status["quantity"], 2)
        self.assertEqual(status["filled_qty"], 2)
        self.assertEqual(status["side"], "sell")
        self.assertEqual(status["time_in_force"], "gtc")
        self.assertEqual(status["status"], "filled")

    def test_sell(self):
        directory = pathlib.Path(__file__).parent.resolve()
        dummy = PaperBroker(str(directory) + "/../dummy_account.yaml")
        interval = {"A": {"interval": Interval.MIN_1, "aggregations": []}}
        dummy.setup(interval)
        order = dummy.sell("A", 2)
        self.assertEqual(order["type"], "STOCK")
        self.assertEqual(order["id"], 0)
        self.assertEqual(order["symbol"], "A")

        status = dummy.fetch_stock_order_status(order["id"])
        self.assertEqual(status["id"], 0)
        self.assertEqual(status["symbol"], "A")
        self.assertEqual(status["quantity"], 2)
        self.assertEqual(status["filled_qty"], 2)
        self.assertEqual(status["side"], "sell")
        self.assertEqual(status["time_in_force"], "gtc")
        self.assertEqual(status["status"], "filled")

    def test_order_option_limit(self):
        dummy = PaperBroker()
        interval = {"A": {"interval": Interval.MIN_1, "aggregations": []}}
        dummy.setup(interval)
        exp_date = dt.datetime(2021, 11, 14) + dt.timedelta(hours=5)
        order = dummy.order_option_limit(
            "buy", "A", 5, 50000, "OPTION", exp_date, 50001
        )
        self.assertEqual(order["type"], "OPTION")
        self.assertEqual(order["id"], 0)
        self.assertEqual(order["symbol"], "A     211114P50001000")

        status = dummy.fetch_option_order_status(order["id"])
        self.assertEqual(status["symbol"], "A     211114P50001000")
        self.assertEqual(status["quantity"], 5)

    def test_commission(self):
        commission_fee = {"buy": 5.76, "sell": "2%"}

        dummy = PaperBroker(commission_fee=commission_fee)
        total_cost = dummy.apply_commission(50, dummy.commission_fee, "buy")
        self.assertEqual(total_cost, 55.76)
        total_cost = dummy.apply_commission(50, dummy.commission_fee, "sell")
        self.assertEqual(total_cost, 49)


if __name__ == "__main__":
    unittest.main()



# Builtins
import io
import unittest
import unittest.mock

from harvest.wizard import Wizard


class TestWizard(unittest.TestCase):
    def test_init(self):
        w = Wizard()

        self.assertEqual(w.text_counter, 0)
        self.assertEqual(w.prompt_counter, 0)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_print(self, mock_stdout):
        w = Wizard()
        w.print("Hello")
        self.assertEqual(mock_stdout.getvalue(), "💬 Hello")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_println(self, mock_stdout):
        w = Wizard()
        w.println("Hello")
        self.assertEqual(mock_stdout.getvalue(), "💬 Hello\n")

    @unittest.mock.patch("builtins.input", return_value="yes")
    def test_get_bool(self, mock_stdout):
        w = Wizard()
        b = w.get_bool()
        self.assertTrue(b)

    @unittest.mock.patch("builtins.input", return_value="hello")
    def test_get_string(self, mock_stdout):
        w = Wizard()
        s = w.get_string()
        self.assertEqual(s, "hello")

    @unittest.mock.patch("builtins.input", return_value="5")
    def test_get_int(self, mock_stdout):
        w = Wizard()
        i = w.get_int()
        self.assertEqual(i, 5)

    @unittest.mock.patch("builtins.input", return_value="")
    def test_wait(self, mock_stdout):
        w = Wizard()
        w.wait_for_input()
        self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()



# Builtin Imports
import re
import sys
import time
import random
import logging
import datetime as dt
from datetime import datetime, timezone as tz
from enum import IntEnum, auto
from zoneinfo import ZoneInfo

# External Imports
import pandas as pd

# Configure a logger used by all of Harvest.
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s : %(name)s : %(levelname)s : %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p",
    handlers=[logging.FileHandler("harvest.log"), logging.StreamHandler(sys.stdout)],
)
debugger = logging.getLogger("harvest")


class Interval(IntEnum):
    SEC_15 = auto()
    MIN_1 = auto()
    MIN_5 = auto()
    MIN_15 = auto()
    MIN_30 = auto()
    HR_1 = auto()
    DAY_1 = auto()


def interval_string_to_enum(str_interval: str):
    if str_interval == "15SEC":
        return Interval.SEC_15
    elif str_interval == "1MIN":
        return Interval.MIN_1
    elif str_interval == "5MIN":
        return Interval.MIN_5
    elif str_interval == "15MIN":
        return Interval.MIN_15
    elif str_interval == "30MIN":
        return Interval.MIN_30
    elif str_interval == "1HR":
        return Interval.HR_1
    elif str_interval == "1DAY":
        return Interval.DAY_1
    else:
        raise ValueError(f"Invalid interval string {str_interval}")


def interval_enum_to_string(enum):
    try:
        name = enum.name
        unit, val = name.split("_")
        return val + unit
    except:
        return str(enum)


def is_freq(time, interval):
    """Determine if algorithm should be invoked for the
    current time, given the interval. For example, if interval is 30MIN,
    algorithm should be called when minutes are 0 and 30, like 11:30 or 12:00.
    """
    time = time.astimezone(tz.utc)

    if interval == Interval.MIN_1:
        return True

    minutes = time.minute
    hours = time.hour
    if interval == Interval.DAY_1:
        # TODO: Use API to get real-time market hours
        return minutes == 50 and hours == 19
    elif interval == Interval.HR_1:
        return minutes == 0
    val, _ = expand_interval(interval)

    return minutes % val == 0


def expand_interval(interval: Interval):
    """Given a IntEnum interval, returns the unit of time and the number of units."""
    string = interval.name
    unit, value = string.split("_")
    return int(value), unit


def expand_string_interval(interval: str):
    """Given a string interval, returns the unit of time and the number of units.
    For example, "3DAY" should return (3, "DAY")
    """
    num = [c for c in interval if c.isdigit()]
    value = int("".join(num))
    unit = interval[len(num) :]
    return value, unit


def interval_to_timedelta(interval: Interval) -> dt.timedelta:
    """Converts an IntEnum interval into a timedelta object of equal value."""
    expanded_units = {"DAY": "days", "HR": "hours", "MIN": "minutes"}
    value, unit = expand_interval(interval)
    params = {expanded_units[unit]: value}
    return dt.timedelta(**params)


def symbol_type(symbol):
    """Determines the type of the asset the symbol represents.
    This can be 'STOCK', 'CRYPTO', or 'OPTION'
    """
    if len(symbol) > 6:
        return "OPTION"
    elif symbol[0] == "@":
        return "CRYPTO"
    else:
        return "STOCK"


# =========== DataFrame utils ===========


def normalize_pandas_dt_index(df: pd.DataFrame) -> pd.Index:
    return df.index.floor("min")


def aggregate_df(df, interval: Interval) -> pd.DataFrame:
    sym = df.columns[0][0]
    df = df[sym]
    op_dict = {
        "open": "first",
        "high": "max",
        "low": "min",
        "close": "last",
        "volume": "sum",
    }
    val, unit = expand_interval(interval)
    val = str(val)
    if unit == "1HR":
        val = "H"
    elif unit == "MIN":
        val += "T"
    else:
        val = "D"
    df = df.resample(val).agg(op_dict)
    df.columns = pd.MultiIndex.from_product([[sym], df.columns])

    return df.dropna()


# ========== Date utils ==========


def now() -> dt.datetime:
    """
    Returns the current time precise to the minute in the UTC timezone
    """
    return dt.datetime.now(tz.utc).replace(microsecond=0, second=0)


def epoch_zero() -> dt.datetime:
    """
    Returns a datetime object corresponding to midnight 1/1/1970 UTC
    """
    return dt.datetime(1970, 1, 1, tzinfo=tz.utc)


def date_to_str(day) -> str:
    return day.strftime("%Y-%m-%d")


def str_to_date(day) -> str:
    return dt.datetime.strptime(day, "%Y-%m-%d")


def str_to_datetime(date: str) -> dt.datetime:
    """
    :date: A string in the format YYYY-MM-DD hh:mm
    """
    if len(date) <= 10:
        return dt.datetime.strptime(date, "%Y-%m-%d")
    return dt.datetime.strptime(date, "%Y-%m-%d %H:%M")


def convert_input_to_datetime(datetime, timezone: ZoneInfo):

    if datetime is None:
        return None
    elif isinstance(datetime, Timestamp):
        datetime = tz.localize(datetime.timestamp)
    elif isinstance(datetime, str):
        datetime = str_to_datetime(datetime)
    elif isinstance(datetime, dt.datetime):
        datetime = datetime.replace(tzinfo=timezone)
    else:
        raise ValueError(f"Cannot convert {datetime} to datetime.")

    datetime = datetime.replace(tzinfo=timezone)
    datetime = datetime.astimezone(tz.utc)

    return datetime


def convert_input_to_timedelta(period):
    """Converts period into a timedelta object.
    Period can be a string, timedelta object, or a Timerange object."""
    if period is None:
        return None
    elif isinstance(period, Timerange):
        return period.timerange
    elif isinstance(period, str):
        expanded_units = {"DAY": "days", "HR": "hours", "MIN": "minutes"}
        val, unit = expand_string_interval(period)
        return dt.timedelta(**{expanded_units[unit]: val})
    elif isinstance(period, dt.timedelta):
        return period
    else:
        raise ValueError(f"Cannot convert {period} to timedelta.")


def has_timezone(date: dt.datetime) -> bool:
    return date.tzinfo is not None and date.tzinfo.utcoffset(date) is not None


def pandas_timestamp_to_local(df: pd.DataFrame, timezone: ZoneInfo) -> pd.DataFrame:
    """
    Converts the timestamp of a dataframe to local time, represented as a
    timezone naive datetime object.
    """
    df.index = df.index.map(lambda x: datetime_utc_to_local(x, timezone))
    return df


def pandas_datetime_to_utc(df: pd.DataFrame, timezone: ZoneInfo) -> pd.DataFrame:
    """
    Converts timezone naive datetime index of dataframes to a timezone aware datetime index
    adjusted to UTC timezone.
    """
    df.index = df.index.map(lambda x: x.replace(tzinfo=timezone).astimezone(tz.utc))
    return df


def datetime_utc_to_local(date_time: dt.datetime, timezone: ZoneInfo) -> dt.datetime:
    """
    Converts a datetime object in UTC to local time, represented as a
    timezone naive datetime object.
    """
    date_time = date_time.to_pydatetime()
    new_tz = date_time.astimezone(timezone)
    return new_tz.replace(tzinfo=None)


class Timestamp:
    def __init__(self, *args) -> None:
        if len(args) == 1:
            timestamp = args[0]
            if isinstance(timestamp, str):
                self.timestamp = str_to_datetime(timestamp)
            elif isinstance(timestamp, dt.datetime):
                self.timestamp = timestamp
            else:
                raise ValueError(f"Invalid timestamp type {type(timestamp)}")
        elif len(args) > 1:
            self.timestamp = dt.datetime(*args)

    def __sub__(self, other):
        return Timerange(self.timestamp - other.timestamp)


class Timerange:
    def __init__(self, *args) -> None:
        if len(args) == 1:
            timerange = args[1]
            if isinstance(timerange, dt.timedelta):
                self.timerange = timerange
            else:
                raise ValueError(f"Invalid timestamp type {type(timerange)}")
        elif len(args) > 1:
            range_list = ["days", "hours", "minutes"]
            dict = {range_list[i]: arg for i, arg in enumerate(args)}
            self.timerange = dt.timedelta(**dict)


# ========== Misc. utils ==========
def mark_up(x):
    return round(x * 1.05, 2)


def mark_down(x):
    return round(x * 0.95, 2)


def is_crypto(symbol: str) -> bool:
    return symbol[0] == "@"


############ Functions used for testing #################


def gen_data(symbol: str, points: int = 50) -> pd.DataFrame:
    n = now()
    index = [n - dt.timedelta(minutes=1) * i for i in range(points)][::-1]
    df = pd.DataFrame(index=index, columns=["low", "high", "close", "open", "volume"])
    df.index.rename("timestamp", inplace=True)
    df["low"] = [random.random() for _ in range(points)]
    df["high"] = [random.random() for _ in range(points)]
    df["close"] = [random.random() for _ in range(points)]
    df["open"] = [random.random() for _ in range(points)]
    df["volume"] = [random.random() for _ in range(points)]
    # df.index = normalize_pandas_dt_index(df)
    df.columns = pd.MultiIndex.from_product([[symbol], df.columns])

    return df


def not_gh_action(func):
    def wrapper(*args, **kwargs):
        if "GITHUB_ACTIONS" in os.environ:
            return
        func(*args, **kwargs)
        return wrapper



import pandas as pd

from harvest.plugin._base import Plugin


class DoltOptionsPlugin:
    """
    Interfaces with the Dolt CLI to get data from the
    `https://www.dolthub.com/repositories/post-no-preference/options`
    database.
    """

    def __init__(self, name: str = "dolt_options", path: str = "options"):
        """
        :path: The path to the post-no-preference/options repo.
        """
        super.__init__(name, ["doltpy"])
        from doltpy.cli import Dolt
        from doltpy.cli import read

        self.dolt = Dolt(path)
        self.read = read

    def installation(self) -> str:
        """
        Returns how to install the necessary prerequsites for this plugin.
        """
        return """
        Useful Links:
        https://www.dolthub.com/
        https://github.com/dolthub/doltpy

        Install dolt (*nix):
        sudo curl -L https://github.com/dolthub/dolt/releases/latest/download/install.sh | sudo bash

        Get post-no-preference/options repo:
        dolt clone post-no-preference/options

        Install doltpy:
        pip install doltpy
        """

    # -------------- Plugin specific methods -------------- #

    def query(self, query: str) -> pd.DataFrame:
        return self.read.read_pandas_sql(self.dolt, query)



# Builtins
import re
import datetime as dt
from typing import Any, Dict, List, Tuple

# External libraries
import pandas as pd
import yaml

# Submodule imports
from harvest.api._base import API
from harvest.api.dummy import DummyStreamer
from harvest.utils import *


class PaperBroker(API):
    """DummyBroker, as its name implies, is a dummy broker class that can
    be useful for testing algorithms. When used as a streamer, it will return
    randomly generated prices. When used as a broker, it paper trades.
    """

    interval_list = [
        Interval.MIN_1,
        Interval.MIN_5,
        Interval.MIN_15,
        Interval.MIN_30,
        Interval.HR_1,
        Interval.DAY_1,
    ]

    def __init__(self, account_path: str = None, commission_fee=0, streamer=None):
        """
        :commission_fee: When this is a number it is assumed to be a flat price
            on all buys and sells of assets. When this is a string formatted as
            'XX%' then it is assumed that commission fees are that percent of the
            original cost of the buy or sell. When commission fee is a dictionary
            with the keys 'buy' and 'sell' you can specify different commission
            fees when buying and selling assets. The values must be numbers or
            strings formatted as 'XX%'.
        """

        self.stocks = []
        self.options = []
        self.cryptos = []
        self.orders = []

        self.equity = 1000000.0
        self.cash = 1000000.0
        self.buying_power = 1000000.0
        self.multiplier = 1
        self.commission_fee = commission_fee
        self.id = 0
        self.streamer = DummyStreamer() if streamer is None else streamer

        if account_path:
            with open(account_path, "r") as f:
                account = yaml.safe_load(f)
                self.equity = account["equity"]
                self.cash = account["cash"]
                self.buying_power = account["buying_power"]
                self.multiplier = account["multiplier"]

                for stock in account["stocks"]:
                    self.stocks.append(stock)

                for crypto in account["cryptos"]:
                    self.cryptos.append(crypto)

    def setup(self, interval, trader_main=None):
        super().setup(interval, trader_main)

    # -------------- Streamer methods -------------- #

    # Not implemented:
    #   fetch_price_history
    #   fetch_chain_info
    #   fetch_chain_data
    #   fetch_option_market_data

    # ------------- Broker methods ------------- #

    def fetch_stock_positions(self) -> List[Dict[str, Any]]:
        return self.stocks

    def fetch_option_positions(self) -> List[Dict[str, Any]]:
        return self.options

    def fetch_crypto_positions(self) -> List[Dict[str, Any]]:
        return self.cryptos

    def update_option_positions(self, positions) -> List[Dict[str, Any]]:
        for r in self.options:
            occ_sym = r["symbol"]
            price = self.streamer.fetch_option_market_data(occ_sym)["price"]

            r["current_price"] = price
            r["market_value"] = price * r["quantity"] * 100
            r["cost_basis"] = r["avg_price"] * r["quantity"] * 100

    def fetch_account(self) -> Dict[str, Any]:
        self.equity = self._calc_equity()
        return {
            "equity": self.equity,
            "cash": self.cash,
            "buying_power": self.buying_power,
            "multiplier": self.multiplier,
        }

    def fetch_stock_order_status(self, id: int) -> Dict[str, Any]:
        ret = next(r for r in self.orders if r["id"] == id)
        sym = ret["symbol"]

        price = self.streamer.fetch_price_history(
            sym,
            self.interval[sym]["interval"],
            self.streamer.get_current_time() - dt.timedelta(days=7),
            self.streamer.get_current_time(),
        )[sym]["close"][-1]

        qty = ret["quantity"]
        original_price = price * qty
        # If order is open, simulate asset buy/sell if possible
        if ret["status"] == "open":
            lst = self.cryptos if is_crypto(ret["symbol"]) else self.stocks
            pos = next((r for r in lst if r["symbol"] == sym), None)
            if ret["side"] == "buy":
                # Check to see if user has enough funds to buy the stock
                actual_price = self.apply_commission(
                    original_price, self.commission_fee, "sell"
                )
                if self.buying_power < actual_price:
                    debugger.error(
                        f"""Not enough buying power.\n Total price ({actual_price}) exceeds buying power {self.buying_power}.\n Reduce purchase quantity or increase buying power."""
                    )
                elif ret["limit_price"] < price:
                    limit_price = ret["limit_price"]
                    debugger.info(
                        f"Limit price for {sym} is less than current price ({limit_price} < {price})."
                    )
                else:
                    # If asset already exists, buy more. If not, add a new entry
                    if pos is None:
                        lst.append({"symbol": sym, "avg_price": price, "quantity": qty})
                    else:
                        pos["avg_price"] = (
                            pos["avg_price"] * pos["quantity"] + price * qty
                        ) / (qty + pos["quantity"])
                        pos["quantity"] = pos["quantity"] + qty

                    self.cash -= actual_price
                    self.buying_power -= actual_price
                    ret_1 = ret.copy()
                    self.orders.remove(ret)
                    ret = ret_1
                    ret["status"] = "filled"
                    ret["filled_time"] = self.streamer.get_current_time()
                    ret["filled_price"] = price
            else:
                if pos is None:
                    raise Exception(f"Cannot sell {sym}, is not owned")

                pos["quantity"] = pos["quantity"] - qty
                if pos["quantity"] < 1e-8:
                    lst.remove(pos)
                actual_worth = self.apply_commission(
                    original_price, self.commission_fee, "sell"
                )
                self.cash += actual_worth
                self.buying_power += actual_worth
                ret_1 = ret.copy()
                self.orders.remove(ret)
                ret = ret_1
                ret["status"] = "filled"
                ret["filled_time"] = self.streamer.get_current_time()
                ret["filled_price"] = price

            self.equity = self._calc_equity()

        debugger.debug(f"Returning status: {ret}")
        debugger.debug(f"Positions:\n{self.stocks}\n=========\n{self.cryptos}")
        debugger.debug(f"Equity:{self._calc_equity()}")

        return ret

    def fetch_option_order_status(self, id: int) -> Dict[str, Any]:
        ret = next(r for r in self.orders if r["id"] == id)
        sym = ret["base_symbol"]
        occ_sym = ret["symbol"]

        price = self.streamer.fetch_option_market_data(occ_sym)["price"]

        qty = ret["quantity"]
        original_price = price * qty
        # If order has been opened, simulate asset buy/sell
        if ret["status"] == "open":
            pos = next((r for r in self.options if r["symbol"] == occ_sym), None)
            if ret["side"] == "buy":
                # Check to see if user has enough funds to buy the stock
                actual_price = self.apply_commission(
                    original_price, self.commission_fee, "buy"
                )
                if self.buying_power < actual_price:
                    debugger.error(
                        f"""Not enough buying power.\n Total price ({actual_price}) exceeds buying power {self.buying_power}.\n Reduce purchase quantity or increase buying power."""
                    )
                elif ret["limit_price"] < price:
                    limit_price = ret["limit_price"]
                    debugger.info(
                        f"Limit price for {sym} is less than current price ({limit_price} < {price})."
                    )
                else:
                    # If asset already exists, buy more. If not, add a new entry
                    if pos is None:
                        sym, date, option_type, strike = self.occ_to_data(occ_sym)
                        self.options.append(
                            {
                                "base_symbol": sym,
                                "symbol": ret["symbol"],
                                "avg_price": price,
                                "quantity": ret["quantity"],
                                "multiplier": 100,
                                "exp_date": date,
                                "strike_price": strike,
                                "type": option_type,
                            }
                        )
                    else:
                        pos["avg_price"] = (
                            pos["avg_price"] * pos["quantity"] + price * qty
                        ) / (qty + pos["quantity"])
                        pos["quantity"] = pos["quantity"] + qty

                    self.cash -= actual_price
                    self.buying_power -= actual_price
                    ret["status"] = "filled"
                    ret["filled_time"] = self.streamer.get_current_time()
                    ret["filled_price"] = price
                    debugger.debug(f"After BUY: {self.buying_power}")
                    ret_1 = ret.copy()
                    self.orders.remove(ret)
                    ret = ret_1
            else:
                if pos is None:
                    raise Exception(f"Cannot sell {sym}, is not owned")
                pos["quantity"] = pos["quantity"] - qty
                debugger.debug(f"current:{self.buying_power}")
                actual_price = self.apply_commission(
                    original_price, self.commission_fee, "sell"
                )
                self.cash += actual_price
                self.buying_power += actual_price
                debugger.debug(
                    f"Made {sym} {occ_sym} {qty} {price}: {self.buying_power}"
                )
                if pos["quantity"] < 1e-8:
                    self.options.remove(pos)
                ret["status"] = "filled"
                ret["filled_time"] = self.streamer.get_current_time()
                ret["filled_price"] = price
                ret_1 = ret.copy()
                self.orders.remove(ret)
                ret = ret_1

            self.equity = self._calc_equity()

        debugger.debug(f"Returning status: {ret}")
        debugger.debug(f"Positions:\n{self.stocks}\n{self.options}\n{self.cryptos}")
        debugger.debug(f"Equity:{self._calc_equity()}")

        return ret

    def fetch_crypto_order_status(self, id: int) -> Dict[str, Any]:
        return self.fetch_stock_order_status(id)

    def fetch_order_queue(self) -> List[Dict[str, Any]]:
        return self.orders

    # --------------- Methods for Trading --------------- #

    def order_stock_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        data = {
            "type": "STOCK",
            "symbol": symbol,
            "quantity": quantity,
            "filled_qty": quantity,
            "limit_price": limit_price,
            "id": self.id,
            "time_in_force": in_force,
            "status": "open",
            "side": side,
        }

        self.orders.append(data)
        self.id += 1
        ret = {"type": "STOCK", "id": data["id"], "symbol": data["symbol"]}
        return ret

    def order_crypto_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        data = {
            "type": "CRYPTO",
            "symbol": "@" + symbol,
            "quantity": quantity,
            "filled_qty": quantity,
            "limit_price": limit_price,
            "id": self.id,
            "time_in_force": in_force,
            "status": "open",
            "side": side,
        }

        self.orders.append(data)
        self.id += 1
        ret = {"type": "CRYPTO", "id": data["id"], "symbol": data["symbol"]}
        return ret

    def order_option_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        type: str,
        exp_date: dt.datetime,
        strike: float,
        in_force: str = "gtc",
    ):

        data = {
            "type": "OPTION",
            "symbol": self.data_to_occ(symbol, exp_date, type, strike),
            "quantity": quantity,
            "filled_qty": 0,
            "id": self.id,
            "time_in_force": in_force,
            "status": "open",
            "side": side,
            "limit_price": limit_price,
            "base_symbol": symbol,
        }

        self.orders.append(data)
        self.id += 1
        return {"type": "OPTION", "id": data["id"], "symbol": data["symbol"]}

    # ------------- Helper methods ------------- #

    def _calc_equity(self):
        """
        Calculates the total worth of the broker by adding together the
        worth of all stocks, cryptos, options and cash in the broker.
        """
        e = 0
        for asset in self.stocks + self.cryptos + self.options:
            add = asset["avg_price"] * asset["quantity"]
            if "multiplier" in asset:
                add = add * asset["multiplier"]
            e += add
        e += self.cash
        return e

    def apply_commission(self, inital_price: float, commission_fee, side: str) -> float:
        if side == "buy":
            f = lambda a, b: a + b
        elif side == "sell":
            f = lambda a, b: a - b

        if type(commission_fee) in (int, float):
            return f(inital_price, commission_fee)
        elif type(commission_fee) is str:
            pattern = r"([0-9]+\.?[0-9]*)\%"
            match = re.fullmatch(pattern, commission_fee)
            if match is not None:
                commission_fee = inital_price * 0.01 * float(match.group(1))
                return f(inital_price, commission_fee)
            raise Exception(
                f"`commission_fee` {commission_fee} not valid, must match this regex expression: {pattern}"
            )
        elif type(commission_fee) is dict:
            return self.apply_commission(inital_price, commission_fee[side], side)



# Builtins
import pathlib
import unittest
import datetime as dt

import pytz
from pandas.testing import assert_frame_equal

from harvest.api.dummy import DummyStreamer
from harvest.utils import *


class TestDummyStreamer(unittest.TestCase):
    def test_fetch_prices(self):
        streamer = DummyStreamer()
        streamer.poll_interval = Interval.MIN_1
        # Get per-minute data
        df_1 = streamer.fetch_price_history("A", Interval.MIN_1)["A"]
        # Check that the last date is 1/1/2020-00:00:00
        self.assertEqual(
            df_1.index[-1], pytz.utc.localize(dt.datetime(2000, 1, 1, 0, 0))
        )
        # Advance the time by 1 minute
        streamer.tick()
        df_2 = streamer.fetch_price_history("A", Interval.MIN_1)["A"]
        # Check that the last date is 1/1/2020-00:01:00
        self.assertEqual(
            df_2.index[-1], pytz.utc.localize(dt.datetime(2000, 1, 1, 0, 1))
        )

        # Check that the two dataframes are the same
        assert_frame_equal(df_1.iloc[1:], df_2.iloc[:-1])

    def test_setup(self):
        dummy = DummyStreamer()
        interval = {
            "A": {"interval": Interval.MIN_1, "agg,regations": []},
            "B": {"interval": Interval.MIN_1, "agg,regations": []},
            "C": {"interval": Interval.MIN_1, "agg,regations": []},
            "@D": {"interval": Interval.MIN_1, "agg,regations": []},
        }
        dummy.setup(interval)

        self.assertEqual(dummy.interval, interval)

    def test_get_stock_price(self):
        dummy = DummyStreamer()
        interval = {
            "A": {"interval": Interval.MIN_1, "agg,regations": []},
            "B": {"interval": Interval.MIN_1, "agg,regations": []},
            "C": {"interval": Interval.MIN_1, "agg,regations": []},
            "@D": {"interval": Interval.MIN_1, "agg,regations": []},
        }
        dummy.setup(interval)
        d = dummy.fetch_latest_stock_price()
        self.assertEqual(len(d), 3)

    def test_get_crypto_price(self):
        dummy = DummyStreamer()
        interval = {
            "A": {"interval": Interval.MIN_1, "agg,regations": []},
            "B": {"interval": Interval.MIN_1, "agg,regations": []},
            "C": {"interval": Interval.MIN_1, "agg,regations": []},
            "@D": {"interval": Interval.MIN_1, "agg,regations": []},
        }
        dummy.setup(interval)
        d = dummy.fetch_latest_crypto_price()
        self.assertTrue("@D" in d)
        self.assertEqual(d["@D"].shape, (1, 5))

    def test_simple_static(self):
        dummy1 = DummyStreamer()
        dummy2 = DummyStreamer()

        df1 = dummy1.fetch_price_history("A", Interval.MIN_1)
        df2 = dummy2.fetch_price_history("A", Interval.MIN_1)

        assert_frame_equal(df1, df2)

    def test_complex_static(self):
        dummy1 = DummyStreamer()
        dummy2 = DummyStreamer()

        df1B = dummy1.fetch_price_history("B", Interval.MIN_5)
        df2A = dummy2.fetch_price_history("A", Interval.MIN_5)
        df1A = dummy1.fetch_price_history("A", Interval.MIN_5)
        df2B = dummy2.fetch_price_history("B", Interval.MIN_5)

        assert_frame_equal(df1A, df2A)
        assert_frame_equal(df1B, df2B)


if __name__ == "__main__":
    unittest.main()



# Builtins
from harvest.api.dummy import DummyStreamer
import pathlib
import unittest
import datetime as dt
import os

import pandas as pd


from harvest.algo import BaseAlgo
from harvest.api._base import API, StreamAPI
from harvest.api.dummy import DummyStreamer
from harvest.trader import PaperTrader
from harvest.utils import *


class TestAPI(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        with open("secret.yaml", "a") as f:
            f.write("data: 0")
            f.close()

    def test_timeout(self):
        stream = StreamAPI()
        stream.fetch_account = lambda: None
        stream.fetch_price_history = lambda x, y, z: pd.DataFrame()
        stream.fetch_account = lambda: {"cash": 100, "equity": 100}
        t = PaperTrader(stream)
        stream.trader = t
        stream.trader_main = t.main
        t.set_symbol(["A", "B"])

        t.start("1MIN", sync=False)

        # Save dummy data
        data = gen_data("A", 10)
        t.storage.store("A", Interval.MIN_1, data)
        data = gen_data("B", 10)
        t.storage.store("B", Interval.MIN_1, data)

        # Save the last datapoint of B
        a_cur = t.storage.load("A", Interval.MIN_1)
        b_cur = t.storage.load("B", Interval.MIN_1)

        # Only send data for A
        data = gen_data("A", 1)
        data.index = [stream.timestamp + dt.timedelta(minutes=1)]
        data = {"A": data}
        stream.main(data)

        # Wait for the timeout
        time.sleep(2)

        # Check if A has been added to storage
        self.assertEqual(
            a_cur["A"]["close"][-1],
            t.storage.load("A", Interval.MIN_1)["A"]["close"][-2],
        )
        self.assertEqual(
            data["A"]["A"]["close"][-1],
            t.storage.load("A", Interval.MIN_1)["A"]["close"][-1],
        )
        # Check if B has been set to 0
        self.assertEqual(
            b_cur["B"]["close"][-1],
            t.storage.load("B", Interval.MIN_1)["B"]["close"][-2],
        )
        self.assertEqual(
            0,
            t.storage.load("B", Interval.MIN_1)["B"]["close"][-1],
        )

    def test_timeout_cancel(self):
        stream = StreamAPI()
        stream.fetch_account = lambda: None
        stream.fetch_price_history = lambda x, y, z: pd.DataFrame()
        stream.fetch_account = lambda: {"cash": 100, "equity": 100}
        t = PaperTrader(stream)
        t.set_algo(BaseAlgo())
        stream.trader = t
        stream.trader_main = t.main
        t.set_symbol(["A", "B"])

        t.start("1MIN", sync=False)

        # Save dummy data
        data = gen_data("A", 10)
        t.storage.store("A", Interval.MIN_1, data)
        data = gen_data("B", 10)
        t.storage.store("B", Interval.MIN_1, data)

        # Save the last datapoint of B
        a_cur = t.storage.load("A", Interval.MIN_1)
        b_cur = t.storage.load("B", Interval.MIN_1)

        # Send data for A and B
        data_a = gen_data("A", 1)
        data_a.index = [a_cur.index[-1] + dt.timedelta(minutes=1)]
        data_a = {"A": data_a}
        data_b = gen_data("B", 1)
        data_b.index = [b_cur.index[-1] + dt.timedelta(minutes=1)]
        data_b = {"B": data_b}
        stream.main(data_a)

        # Wait
        time.sleep(0.1)
        stream.main(data_b)

        # Check if A has been added to storage
        self.assertEqual(
            a_cur["A"]["close"][-1],
            t.storage.load("A", Interval.MIN_1)["A"]["close"][-2],
        )
        self.assertEqual(
            data_a["A"]["A"]["close"][-1],
            t.storage.load("A", Interval.MIN_1)["A"]["close"][-1],
        )
        # Check if B has been added to storage
        self.assertEqual(
            b_cur["B"]["close"][-1],
            t.storage.load("B", Interval.MIN_1)["B"]["close"][-2],
        )
        self.assertEqual(
            data_b["B"]["B"]["close"][-1],
            t.storage.load("B", Interval.MIN_1)["B"]["close"][-1],
        )

    def test_exceptions(self):
        api = API()

        self.assertEqual(api.create_secret("I dont exists"), False)

        try:
            api.fetch_price_history('A', Interval.MIN_1, now(), now())
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this streamer method: `fetch_price_history`.")

        try:
            api.fetch_chain_info("A")
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this streamer method: `fetch_chain_info`.")

        try:
            api.fetch_chain_data("A")
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this streamer method: `fetch_chain_data`.")

        try:
            api.fetch_option_market_data("A")
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this streamer method: `fetch_option_market_data`.")

        try:
            api.fetch_account()
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this broker method: `fetch_account`.")

        try:
            api.fetch_stock_order_status(0)
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this broker method: `fetch_stock_order_status`.")

        try:
            api.fetch_option_order_status(0)
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this broker method: `fetch_option_order_status`.")

        try:
            api.fetch_crypto_order_status(0)
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this broker method: `fetch_crypto_order_status`.")

        try:
            api.order_stock_limit("buy", "A", 5, 7)
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this broker method: `order_stock_limit`.")

        try:
            api.order_crypto_limit("buy", "@A", 5, 7)
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this broker method: `order_crypto_limit`.")

        try:
            api.order_option_limit("buy", "A", 5, 7, "call", now(), 8)
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this broker method: `order_option_limit`.")

        try:
            api.buy('A', -1, 0)
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this broker method: `order_stock_limit`.")

        try:
            api.sell('A', -1, 0)
            self.assertTrue(False)
        except NotImplementedError as e:
            self.assertEqual(str(e), "API does not support this broker method: `order_stock_limit`.")

            

    def test_base_cases(self):
        api = API()

        api.refresh_cred()
        api.exit()
        self.assertEqual(api.fetch_stock_positions(), [])
        self.assertEqual(api.fetch_option_positions(), [])
        self.assertEqual(api.fetch_crypto_positions(), [])
        api.update_option_positions([])
        self.assertEqual(api.fetch_order_queue(), [])

    def test_run_once(self):
        api = API()
        fn = lambda x: x + 1
        wrapper = API._run_once(fn)
        self.assertEqual(wrapper(5), 6)
        self.assertTrue(wrapper(5) is None)

    def test_timestamp(self):
        api = API()
        self.assertTrue(now() >= api.current_timestamp())

    @classmethod
    def tearDownClass(self):
        os.remove("secret.yaml")


if __name__ == "__main__":
    unittest.main()



# Builtins
import unittest
import time

# Submodule imports
from harvest.trader import PaperTrader
from harvest.algo import BaseAlgo
from harvest.api.dummy import DummyStreamer
from harvest.api.paper import PaperBroker

# from harvest.api.robinhood import Robinhood

import datetime as dt

from harvest.utils import gen_data


class TestPaperTrader(unittest.TestCase):
    def test_trader_adding_symbol(self):
        t = PaperTrader(DummyStreamer())
        t.set_symbol("A")
        self.assertEqual(t.watchlist_global[0], "A")

    def test_start_do_nothing(self):
        t = PaperTrader(DummyStreamer())
        t.set_symbol("A")
        t.set_algo(BaseAlgo())
        t.start("1MIN")

        self.assertTrue(True)

    def test_no_streamer(self):
        """
        If streamer is not specified, by default
        it should be set to DummyStreamer, and broker set to
        """
        t = PaperTrader(DummyStreamer())

        self.assertIsInstance(t.streamer, DummyStreamer)
        self.assertIsInstance(t.broker, PaperBroker)

    # def test_broker_set(self):
    #     """If a single API class is set, it should be set as
    #     a streamer and a broker"""
    #     t = PaperTrader( Robinhood() )

    #     self.assertIsInstance(t.streamer, Robinhood)
    #     self.assertIsInstance(t.broker, Robinhood)

    def test_dummy_streamer(self):
        """If streamer is DummyStreamer, broker should be PaperBroker"""
        t = PaperTrader(DummyStreamer())

        self.assertIsInstance(t.streamer, DummyStreamer)
        self.assertIsInstance(t.broker, PaperBroker)

    def test_invalid_aggregation(self):
        """If invalid aggregation is set, it should raise an error"""
        t = PaperTrader(DummyStreamer())
        with self.assertRaises(Exception):
            t.start("30MIN", ["5MIN", "1DAY"])


if __name__ == "__main__":
    unittest.main()



from typing import List
from harvest.utils import debugger


class Plugin:
    """
    As the name implies, it implements funtionalities that the Trader class
    alone cannot support, like retrieving earning report dates.

    The name of the plugin will follow the format:
        {Name of source}{Functionality added}Plugin
    """

    def __init__(self, name: str, dependencies: List[str] = []):
        """
        :name: a string that will serve as the name of the plugin. Must
        be a valid python variable because in the algo class, this plugin
        will be an attribute of the algo and accessable via algo.name.
        :dependencies: a list of python packages as strings that will be
        checked to ensure they are installed.
        Put all imports here in order to test if we have them. If not then
        error and call the installation function.
        """
        self.name = name

        for dep in dependencies:
            self._check_dependency(dep)

    def installation(self) -> str:
        """
        Returns how to install the necessary prerequsites for this plugin.
        """
        raise NotImplementedError(
            f"No installation steps for plugin: {type(self).__name__}"
        )

    def _check_dependency(self, dep: str) -> None:
        try:
            exec(f"import {dep}")
        except ModuleNotFoundError as e:
            debugger.error(f"Error importing module!\n {e}")
            debugger.error(self.installation())
            raise e



# Builtins
import shutil
import pathlib
import unittest

import pandas as pd
from pandas.testing import assert_frame_equal

from harvest.storage import CSVStorage
from harvest.utils import *


class TestCSVStorage(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.storage_dir = "test_csv_data"

    def test_create_storage(self):
        storage = CSVStorage(self.storage_dir)

        self.assertEqual(storage.storage_price, {})

    def test_simple_store(self):
        storage = CSVStorage(self.storage_dir)
        data = gen_data("A", 50)
        storage.store("A", Interval.MIN_1, data.copy(True))

        self.assertTrue(not pd.isna(data.iloc[0]["A"]["low"]))
        self.assertTrue("A" in list(storage.storage_price.keys()))
        self.assertEqual(list(storage.storage_price["A"].keys()), [Interval.MIN_1])

    def test_saved_load(self):
        storage1 = CSVStorage(self.storage_dir)
        data = gen_data("B", 50)
        storage1.store("B", Interval.MIN_1, data.copy(True))

        storage2 = CSVStorage(self.storage_dir)
        loaded_data = storage2.load("B", Interval.MIN_1)

        assert_frame_equal(loaded_data, data)

    @classmethod
    def tearDownClass(self):
        path = pathlib.Path(self.storage_dir)
        shutil.rmtree(path)


if __name__ == "__main__":
    unittest.main()



# Builtins
import datetime as dt
from logging import debug
from typing import Any, Dict, List, Tuple

# External libraries
import pandas as pd
import robin_stocks.robinhood as rh
import pytz
import pyotp
import yaml

# Submodule imports
from harvest.api._base import API
from harvest.utils import *


class Robinhood(API):

    interval_list = [Interval.SEC_15, Interval.MIN_5, Interval.HR_1, Interval.DAY_1]
    exchange = "NASDAQ"

    def __init__(self, path=None):
        super().__init__(path)
        self.login()

    def login(self):
        debugger.debug("Logging into Robinhood...")
        totp = pyotp.TOTP(self.config["robin_mfa"]).now()
        rh.login(
            self.config["robin_username"],
            self.config["robin_password"],
            store_session=True,
            mfa_code=totp,
        )

    def refresh_cred(self):
        super().refresh_cred()
        debugger.debug("Logging out of Robinhood...")
        rh.authentication.logout()
        self.login()
        debugger.debug("Logged into Robinhood...")

    # @API._run_once
    def setup(self, interval, trader_main=None):

        super().setup(interval, trader_main)

        # Robinhood only supports 15SEC, 1MIN interval for crypto
        for sym in interval:
            if not is_crypto(sym) and interval[sym]["interval"] < Interval.MIN_5:
                raise Exception(
                    f'Interval {interval[sym]["interval"]} is only supported for crypto'
                )

        # self.__watch_stock = []
        # self.__watch_crypto = []
        # self.__watch_crypto_fmt = []

        # for s in interval:
        #     if is_crypto(s):
        #         self.__watch_crypto_fmt.append(s[1:])
        #         self.__watch_crypto.append(s)
        #     else:
        #         self.__watch_stock.append(s)

        self.__option_cache = {}

    def exit(self):
        self.__option_cache = {}

    # @API._exception_handler
    # def fetch_latest_stock_price(self):
    #     df={}
    #     for s in self.__watch_stock:
    #         ret = rh.get_stock_historicals(
    #             s,
    #             interval=self.__interval_fmt,
    #             span='day',
    #             )
    #         if 'error' in ret or ret == None or (type(ret) == list and len(ret) == 0):
    #             continue
    #         df_tmp = pd.DataFrame.from_dict(ret)
    #         df_tmp = self._format_df(df_tmp, [s], self.interval).iloc[[-1]]
    #         df[s] = df_tmp

    #     return df

    # @API._exception_handler
    # def fetch_latest_crypto_price(self):
    #     df={}
    #     for s in self.__watch_crypto_fmt:
    #         ret = rh.get_crypto_historicals(
    #             s,
    #             interval=self.__interval_fmt,
    #             span='hour',
    #             )
    #         df_tmp = pd.DataFrame.from_dict(ret)
    #         df_tmp = self._format_df(df_tmp, ['@'+s], self.interval).iloc[[-1]]
    #         df['@'+s] = df_tmp

    #     return df

    # -------------- Streamer methods -------------- #

    @API._exception_handler
    def fetch_price_history(
        self,
        symbol: str,
        interval: str,
        start: dt.datetime = None,
        end: dt.datetime = None,
    ):

        if start is None:
            start = dt.datetime(1970, 1, 1)
        if end is None:
            end = dt.datetime.now()

        df = pd.DataFrame()

        if start >= end:
            return df

        get_interval_fmt = ""

        if interval == Interval.SEC_15:  # Not used
            if symbol[0] != "@":
                raise Exception("15SEC interval is only allowed for crypto")
            get_interval_fmt = "15second"
        elif interval == Interval.MIN_1:
            if symbol[0] != "@":
                raise Exception("MIN interval is only allowed for crypto")
            get_interval_fmt = "15second"
        elif interval == Interval.MIN_5:
            get_interval_fmt = "5minute"
        elif interval == Interval.MIN_15:
            get_interval_fmt = "5minute"
        elif interval == Interval.MIN_30:
            get_interval_fmt = "10minute"
        elif interval == Interval.HR_1:
            get_interval_fmt = "hour"
        elif interval == Interval.DAY_1:
            get_interval_fmt = "day"
        else:
            return df

        delta = end - start
        delta = delta.total_seconds()
        delta = delta / 3600
        if interval == Interval.DAY_1 and delta < 24:
            return df
        if delta < 1 and interval in [Interval.SEC_15, Interval.MIN_1]:
            span = "hour"
        elif delta < 24 or interval in [
            Interval.MIN_5,
            Interval.MIN_15,
            Interval.MIN_30,
            Interval.HR_1,
        ]:
            span = "day"
        elif delta < 24 * 28:
            span = "month"
        elif delta < 24 * 300:
            span = "year"
        else:
            span = "5year"

        if symbol[0] == "@":
            ret = rh.get_crypto_historicals(
                symbol[1:], interval=get_interval_fmt, span=span
            )
        else:
            ret = rh.get_stock_historicals(symbol, interval=get_interval_fmt, span=span)

        df = pd.DataFrame.from_dict(ret)
        df = self._format_df(df, [symbol], interval)
        df = aggregate_df(df, interval)

        return df

    @API._exception_handler
    def fetch_chain_info(self, symbol: str):
        ret = rh.get_chains(symbol)
        return {
            "id": "n/a",
            "exp_dates": [str_to_date(s) for s in ret["expiration_dates"]],
            "multiplier": ret["trade_value_multiplier"],
        }

    @API._exception_handler
    def fetch_chain_data(self, symbol: str, date: dt.datetime):

        if (
            bool(self.__option_cache)
            and symbol in self.__option_cache
            and date in self.__option_cache[symbol]
        ):
            return self.__option_cache[symbol][date]

        ret = rh.find_tradable_options(symbol, date_to_str(date))
        exp_date = []
        strike = []
        type = []
        id = []
        occ = []
        for entry in ret:
            date = entry["expiration_date"]
            date = dt.datetime.strptime(date, "%Y-%m-%d")
            date = pytz.utc.localize(date)
            exp_date.append(date)
            price = float(entry["strike_price"])
            strike.append(price)
            type.append(entry["type"])
            id.append(entry["id"])

            occ.append(self.data_to_occ(symbol, date, type[-1], price))

        df = pd.DataFrame(
            {
                "occ_symbol": occ,
                "exp_date": exp_date,
                "strike": strike,
                "type": type,
                "id": id,
            }
        )
        df = df.set_index("occ_symbol")

        if symbol not in self.__option_cache:
            self.__option_cache[symbol] = {}
        self.__option_cache[symbol][date] = df

        return df

    @API._exception_handler
    def fetch_option_market_data(self, symbol: str):

        sym, date, type, price = self.occ_to_data(symbol)
        ret = rh.get_option_market_data(
            sym, date.strftime("%Y-%m-%d"), str(price), type
        )
        ret = ret[0][0]
        return {
            "price": float(ret["adjusted_mark_price"]),
            "ask": float(ret["ask_price"]),
            "bid": float(ret["bid_price"]),
        }

    # ------------- Broker methods ------------- #

    @API._exception_handler
    def fetch_stock_positions(self):
        ret = rh.get_open_stock_positions()
        pos = []
        for r in ret:
            # 0 quantity means the order was not fulfilled yet
            if float(r["quantity"]) < 0.0001:
                continue
            sym = rh.get_symbol_by_url(r["instrument"])
            pos.append(
                {
                    "symbol": sym,
                    "avg_price": float(r["average_buy_price"]),
                    "quantity": float(r["quantity"]),
                }
            )
        return pos

    @API._exception_handler
    def fetch_option_positions(self):
        ret = rh.get_open_option_positions()
        pos = []
        # print(ret)
        for r in ret:
            # Get option data such as expiration date
            data = rh.get_option_instrument_data_by_id(r["option_id"])
            pos.append(
                {
                    "base_symbol": r["chain_symbol"],
                    "avg_price": float(r["average_price"])
                    / float(r["trade_value_multiplier"]),
                    "quantity": float(r["quantity"]),
                    "multiplier": float(r["trade_value_multiplier"]),
                    "exp_date": data["expiration_date"],
                    "strike_price": float(data["strike_price"]),
                    "type": data["type"],
                }
            )
            date = data["expiration_date"]
            date = dt.datetime.strptime(date, "%Y-%m-%d")
            pos[-1]["symbol"] = self.data_to_occ(
                r["chain_symbol"], date, data["type"], float(data["strike_price"])
            )

        return pos

    @API._exception_handler
    def fetch_crypto_positions(self, key=None):
        ret = rh.get_crypto_positions()
        pos = []
        for r in ret:
            # Note: It seems Robinhood misspelled 'cost_bases',
            # it should be 'cost_basis'
            if len(r["cost_bases"]) <= 0:
                continue
            qty = float(r["cost_bases"][0]["direct_quantity"])
            if qty < 1e-10:
                continue

            pos.append(
                {
                    "symbol": "@" + r["currency"]["code"],
                    "avg_price": float(r["cost_bases"][0]["direct_cost_basis"]) / qty,
                    "quantity": qty,
                }
            )
        return pos

    @API._exception_handler
    def update_option_positions(self, positions: List[Any]):
        for r in positions:
            sym, date, type, price = self.occ_to_data(r["symbol"])
            upd = rh.get_option_market_data(
                sym, date.strftime("%Y-%m-%d"), str(price), type
            )
            upd = upd[0][0]
            r["current_price"] = float(upd["adjusted_mark_price"])
            r["market_value"] = float(upd["adjusted_mark_price"]) * r["quantity"]
            r["cost_basis"] = r["avg_price"] * r["quantity"]

    @API._exception_handler
    def fetch_account(self):
        ret = rh.load_phoenix_account()
        ret = {
            "equity": float(ret["equities"]["equity"]["amount"]),
            "cash": float(ret["uninvested_cash"]["amount"]),
            "buying_power": float(ret["account_buying_power"]["amount"]),
            "multiplier": float(-1),
        }
        return ret

    @API._exception_handler
    def fetch_stock_order_status(self, id):
        ret = rh.get_stock_order_info(id)
        # Check if any of the orders were executed
        executions = ret["executions"]
        if len(executions) > 0:
            filled_time = self._rh_datestr_to_datetime(executions[0]["timestamp"])
            filled_time = filled_time.replace(tzinfo=pytz.utc)
            filled_price = float(executions[0]["effective_price"])
        else:
            filled_time = None
            filled_price = None
        return {
            "type": "STOCK",
            "id": ret["id"],
            "symbol": ret["symbol"],
            "quantity": ret["qty"],
            "filled_quantity": ret["filled_qty"],
            "side": ret["side"],
            "time_in_force": ret["time_in_force"],
            "status": ret["status"],
            "filled_time": filled_time,
            "filled_price": filled_price,
        }

    @API._exception_handler
    def fetch_option_order_status(self, id):
        ret = rh.get_option_order_info(id)
        debugger.debug(ret)
        # Check if any of the orders were executed
        executions = ret["legs"][0]["executions"]
        if len(executions) > 0:
            filled_time = self._rh_datestr_to_datetime(executions[0]["timestamp"])
            filled_time = filled_time.replace(tzinfo=pytz.utc)
            filled_price = float(executions[0]["effective_price"])
        else:
            filled_time = None
            filled_price = None
        return {
            "type": "OPTION",
            "id": ret["id"],
            "symbol": ret["chain_symbol"],
            "qty": ret["quantity"],
            "filled_qty": ret["processed_quantity"],
            "side": ret["legs"][0]["side"],
            "time_in_force": ret["time_in_force"],
            "status": ret["state"],
            "filled_time": filled_time,
            "filled_price": filled_price,
        }

    @API._exception_handler
    def fetch_crypto_order_status(self, id):
        ret = rh.get_crypto_order_info(id)
        debugger.debug(ret)
        # Check if any of the orders were executed
        executions = ret["executions"]
        if len(executions) > 0:
            filled_time = self._rh_datestr_to_datetime(executions[0]["timestamp"])
            filled_time = filled_time.replace(tzinfo=pytz.utc)
            filled_price = float(executions[0]["effective_price"])
        else:
            filled_time = None
            filled_price = None

        return {
            "type": "CRYPTO",
            "id": ret["id"],
            "symbol": ret["symbol"],
            "quantity": float(ret["quantity"]),
            "filled_qty": float(ret["cumulative_quantity"]),
            "side": ret["side"],
            "time_in_force": ret["time_in_force"],
            "status": ret["state"],
            "filled_time": filled_time,
            "filled_price": filled_price,
        }

    @API._exception_handler
    def fetch_order_queue(self):
        queue = []
        ret = rh.get_all_open_stock_orders()
        for r in ret:
            sym = rh.get_symbol_by_url(r["instrument"])
            queue.append(
                {
                    "type": "STOCK",
                    "symbol": sym,
                    "quantity": r["quantity"],
                    "filled_qty": r["cumulative_quantity"],
                    "id": r["id"],
                    "time_in_force": r["time_in_force"],
                    "status": r["state"],
                    "side": r["side"],
                    "filled_time": None,
                    "filled_price": None,
                }
            )

        ret = rh.get_all_open_option_orders()
        for r in ret:
            debugger.debug(r)
            legs = [{"id": l["id"], "side": l["side"]} for l in r["legs"]]
            date = r["legs"][0]["expiration_date"]
            date = dt.datetime.strptime(date, "%Y-%m-%d")
            s = self.data_to_occ(
                r["chain_symbol"],
                date,
                r["legs"][0]["option_type"],
                float(r["legs"][0]["strike_price"]),
            )
            queue.append(
                {
                    "type": "OPTION",
                    "symbol": s,
                    "base_symbol": r["chain_symbol"],
                    "quantity": r["quantity"],
                    "filled_qty": r["processed_quantity"],
                    "id": r["id"],
                    "time_in_force": r["time_in_force"],
                    "status": r["state"],
                    "legs": legs,
                    "filled_time": None,
                    "filled_price": None,
                }
            )
        ret = rh.get_all_open_crypto_orders()
        for r in ret:
            sym = rh.get_symbol_by_url(r["instrument"])
            queue.append(
                {
                    "type": "CRYPTO",
                    "symbol": sym,
                    "quantity": r["quantity"],
                    "filled_qty": r["cumulative_quantity"],
                    "id": r["id"],
                    "time_in_force": r["time_in_force"],
                    "status": r["state"],
                    "side": r["side"],
                    "filled_time": None,
                    "filled_price": None,
                }
            )
        return queue

    # --------------- Methods for Trading --------------- #

    # Order functions are not wrapped in the exception handler to prevent duplicate
    # orders from being made.
    def order_stock_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        ret = None
        try:
            if side == "buy":
                ret = rh.order_buy_limit(
                    symbol=symbol,
                    quantity=quantity,
                    timeInForce=in_force,
                    limitPrice=limit_price,
                )
            else:
                ret = rh.order_sell_limit(
                    symbol=symbol,
                    quantity=quantity,
                    timeInForce=in_force,
                    limitPrice=limit_price,
                )
            return {
                "type": "STOCK",
                "id": ret["id"],
                "symbol": symbol,
            }
        except:
            debugger.error("Error while placing order.\nReturned: {ret}", exc_info=True)
            raise Exception("Error while placing order.")

    def order_crypto_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        ret = None
        try:
            if side == "buy":
                ret = rh.order_buy_crypto_limit(
                    symbol=symbol,
                    quantity=quantity,
                    timeInForce=in_force,
                    limitPrice=limit_price,
                )
            else:
                ret = rh.order_sell_crypto_limit(
                    symbol=symbol,
                    quantity=quantity,
                    timeInForce=in_force,
                    limitPrice=limit_price,
                )
            return {
                "type": "CRYPTO",
                "id": ret["id"],
                "symbol": "@" + symbol,
            }
        except:
            debugger.error("Error while placing order.\nReturned: {ret}", exc_info=True)
            raise Exception("Error while placing order.")

    def order_option_limit(
        self,
        side: str,
        symbol: str,
        quantity: int,
        limit_price: float,
        option_type,
        exp_date: dt.datetime,
        strike,
        in_force: str = "gtc",
    ):
        ret = None
        try:
            if side == "buy":
                ret = rh.order_buy_option_limit(
                    positionEffect="open",
                    creditOrDebit="debit",
                    price=limit_price,
                    symbol=symbol,
                    quantity=quantity,
                    expirationDate=exp_date.strftime("%Y-%m-%d"),
                    strike=strike,
                    optionType=option_type,
                    timeInForce=in_force,
                )
            else:
                ret = rh.order_sell_option_limit(
                    positionEffect="close",
                    creditOrDebit="credit",
                    price=limit_price,
                    symbol=symbol,
                    quantity=quantity,
                    expirationDate=exp_date.strftime("%Y-%m-%d"),
                    strike=strike,
                    optionType=option_type,
                    timeInForce=in_force,
                )
            return {
                "type": "OPTION",
                "id": ret["id"],
                "symbol": symbol,
            }
        except:
            debugger.error("Error while placing order.\nReturned: {ret}", exc_info=True)
            raise Exception("Error while placing order")

    def _format_df(
        self, df: pd.DataFrame, watch: List[str], interval: str, latest: bool = False
    ):
        # Robinhood returns offset-aware timestamps based on timezone GMT-0, or UTC
        df["timestamp"] = pd.to_datetime(df["begins_at"])
        df = df.set_index(["timestamp"])
        df = df.drop(["begins_at"], axis=1)
        df = df.rename(
            columns={
                "open_price": "open",
                "close_price": "close",
                "high_price": "high",
                "low_price": "low",
            }
        )
        df = df[["open", "close", "high", "low", "volume"]].astype(float)

        # RH doesn't support 1MIN intervals natively, so it must be
        # aggregated from 15SEC intervals. In such case, datapoints that are not a full minute should be dropped
        # in case too much time has passed since when the API call was made.
        if interval == "1MIN":
            while df.index[-1].second != 45:
                df = df.iloc[:-1]

        df.columns = pd.MultiIndex.from_product([watch, df.columns])

        return df.dropna()

    def _rh_datestr_to_datetime(self, date_str: str):
        date_str = date_str[:-3] + date_str[-2:]
        return dt.datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%S.%f%z")

    def create_secret(self, path):
        import harvest.wizard as wizard

        w = wizard.Wizard()

        w.println(
            "Hmm, looks like you haven't set up login credentials for Robinhood yet."
        )
        should_setup = w.get_bool("Do you want to set it up now?", default="y")

        if not should_setup:
            w.println("You can't use Robinhood unless we can log you in.")
            w.println("You can set up the credentials manually, or use other brokers.")
            return False

        w.println("Alright! Let's get started")

        have_account = w.get_bool("Do you have a Robinhood account?", default="y")
        if not have_account:
            w.println(
                "In that case you'll first need to make an account. I'll wait here, so hit Enter or Return when you've done that."
            )
            w.wait_for_input()

        have_mfa = w.get_bool(
            "Do you have Two Factor Authentication enabled?", default="y"
        )

        if not have_mfa:
            w.println(
                "Robinhood (and Harvest) requires users to have 2FA enabled, so we'll turn that on next."
            )
        else:
            w.println(
                "We'll need to reconfigure 2FA to use Harvest, so temporarily disable 2FA"
            )
            w.wait_for_input()

        w.println(
            "Enable 2FA. Robinhood should ask you what authentication method you want to use."
        )
        w.wait_for_input()
        w.println("Select 'Authenticator App'.")
        w.wait_for_input()
        w.println("Select 'Can't scan'.")
        w.wait_for_input()

        mfa = w.get_string(
            "You should see a string of letters and numbers on the screen. Type it in here",
            pattern=r"[\d\w]+",
        )
        while True:
            try:
                totp = pyotp.TOTP(mfa).now()
            except:
                mfa = w.get_string(
                    "Woah😮 Something went wrong. Make sure you typed in the code correctly.",
                    pattern=r"[\d\w]+",
                )
                continue
            break

        w.print(
            f"Good! Robinhood should now be asking you for a 6-digit passcode. Type in: {totp}"
        )
        w.print(
            f"⚠️  Beware, this passcode expires in a few seconds! If you couldn't type it in time, it should be regenerated."
        )

        new_passcode = True
        while new_passcode:
            new_passcode = w.get_bool(
                "Do you want to generate a new passcode?", default="n"
            )
            if new_passcode:
                totp = pyotp.TOTP(mfa).now()
                w.print(f"New passcode: {totp}")
            else:
                break

        w.println(
            "Robinhood will show you a backup code. This is useful when 2FA fails, so make sure to keep it somewhere safe."
        )
        w.wait_for_input()
        w.println(
            "It is recommended you also set up 2FA using an app like Authy or Google Authenticator, so you don't have to run this setup wizard every time you log into Robinhood."
        )
        w.wait_for_input()
        w.println(
            f"Open an authenticator app of your choice, and use the MFA code you typed in earlier to set up OTP passcodes for Robinhood: {mfa}"
        )
        w.wait_for_input()

        w.println(f"Almost there! Type in your username and password for Robinhood")

        username = w.get_string("Username: ")
        password = w.get_password("Password: ")

        w.println(f"All steps are complete now 🎉. Generating secret.yml...")

        d = {
            "robin_mfa": f"{mfa}",
            "robin_username": f"{username}",
            "robin_password": f"{password}",
        }

        with open(path, "w") as file:
            yml = yaml.dump(d, file)

        w.println(
            f"secret.yml has been created! Make sure you keep this file somewhere secure and never share it with other people."
        )

        return True



# Builtins
import pathlib
import unittest
import datetime as dt

from harvest.api.yahoo import YahooStreamer
from harvest.trader.trader import PaperTrader
from harvest.utils import *


class TestYahooStreamer(unittest.TestCase):
    def test_fetch_prices(self):
        yh = YahooStreamer()
        df = yh.fetch_price_history("SPY", Interval.HR_1)
        df = df["SPY"]
        self.assertEqual(
            list(df.columns.values), ["open", "high", "low", "close", "volume"]
        )

    def test_setup(self):
        yh = YahooStreamer()
        interval = {
            "SPY": {"interval": Interval.MIN_15, "aggregations": []},
            "AAPL": {"interval": Interval.MIN_1, "aggregations": []},
        }

        yh.setup(interval)
        self.assertEqual(yh.poll_interval, Interval.MIN_1)
        self.assertListEqual([s for s in yh.interval], ["SPY", "AAPL"])

    def test_main(self):
        interval = {
            "SPY": {"interval": Interval.MIN_1, "aggregations": []},
            "AAPL": {"interval": Interval.MIN_1, "aggregations": []},
            "@BTC": {"interval": Interval.MIN_1, "aggregations": []},
        }

        def test_main(df):
            self.assertEqual(len(df), 3)
            self.assertEqual(df["SPY"].columns[0][0], "SPY")
            self.assertEqual(df["AAPL"].columns[0][0], "AAPL")
            self.assertEqual(df["@BTC"].columns[0][0], "@BTC")

        yh = YahooStreamer()
        watch = ["SPY", "AAPL", "@BTC"]
        yh.setup(interval, test_main)
        yh.main()

    def test_main_single(self):
        interval = {"SPY": {"interval": Interval.MIN_1, "aggregations": []}}

        def test_main(df):
            self.assertEqual(len(df), 1)
            self.assertEqual(df["SPY"].columns[0][0], "SPY")

        yh = YahooStreamer()
        yh.setup(interval, test_main)
        yh.main()

    def test_chain_info(self):
        t = PaperTrader()
        yh = YahooStreamer()
        interval = {"LMND": {"interval": Interval.MIN_1, "aggregations": []}}
        yh.setup(interval)
        info = yh.fetch_chain_info("LMND")
        self.assertGreater(len(info["exp_dates"]), 0)

    def test_chain_data(self):
        t = PaperTrader()
        yh = YahooStreamer()
        interval = {"LMND": {"interval": Interval.MIN_1, "aggregations": []}}
        yh.setup(interval)
        dates = yh.fetch_chain_info("LMND")["exp_dates"]
        data = yh.fetch_chain_data("LMND", dates[0])
        self.assertGreater(len(data), 0)
        self.assertListEqual(list(data.columns), ["exp_date", "strike", "type"])

        sym = data.index[0]
        df = yh.fetch_option_market_data(sym)

        self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()



from numpy import ERR_CALL
import pandas as pd
import datetime as dt
from threading import Lock
from typing import Tuple
import re

from harvest.utils import *

"""
This module serves as a basic storage system for pandas dataframes in memory.
It is made to be a simplistic interface for the trader class to store various data,
such as stock price history and transaction history of the algorithm.

All implementations of Storage class must store the following data:
- Stock price history: Database of stock prices
    - Symbol: Stock symbol
    - Interval: Interval of the stock price
    - Date: Date of the price, adjusted to UTC timezone
    - Open: Opening price
    - High: Highest price
    - Low: Lowest price
    - Close: Closing price
- Transaction history: Database of *filled* orders
    - Symbol: Symbol of the stock
    - Timestamp: Date of the transaction, adjusted to UTC timezone
    - Algorithm name: Name of the algorithm that made the transaction
    - Side: Buy or sell
    - Quantity: Number of shares
    - Price: Price per share
- Performance History: Database of account equity history,
    for the following intervals:
    - 1 day
    - 1 week
    - 1 month
    - 3 months
    - 1 year, at 1 day intervals
    - all time, at variable intervals
    As for 'all time', the interval will be adjusted as the 
    duration increases.

The exact implementation of these databases is up to the classes that inherit from BaseStorage,
as long as they implement the API properly.
"""


class BaseStorage:
    """
    A basic storage that is thread safe and stores data in memory.
    """

    performance_history_intervals = [
        ("1DAY", 1),
        ("1WEEK", 7),
        ("1MONTH", 30),
        ("3MONTH", 90),
        ("1YEAR", 365),
        ("ALL", None),
    ]

    def __init__(
        self,
        price_storage_size: int = 200,
        price_storage_limit: bool = True,
        transaction_storage_size: int = 200,
        transaction_storage_limit: bool = True,
        performance_storage_size: int = 200,
        performance_storage_limit: bool = True,
    ):
        """
        queue_size: The maximum number of data points to store for asset price history.
            This helps prevent the database from becoming infinitely large as time progresses.
        limit_size: Whether to limit the size of price history to queue_size.
            This may be set to False if the storage is being used for backtesting, in which case
            you would want to store as much data as possible.
        """
        self.storage_lock = Lock()  # Lock

        self.price_storage_size = price_storage_size
        self.price_storage_limit = price_storage_limit
        self.transaction_storage_size = transaction_storage_size
        self.transaction_storage_limit = transaction_storage_limit
        self.performance_storage_size = performance_storage_size
        self.performance_storage_limit = performance_storage_limit

        # BaseStorage uses a python dictionary to store the data,
        # where key is asset symbol and value is a pandas dataframe.
        self.storage_price = {}

        self.storage_transaction = pd.DataFrame(
            columns=[
                "timestamp",
                "algorithm_name",
                "symbol",
                "side",
                "quantity",
                "price",
            ]
        )

        self.storage_performance = {}
        for interval, _ in self.performance_history_intervals:
            self.storage_performance[interval] = pd.DataFrame(
                columns=["equity"], index=[]
            )

    def store(
        self, symbol: str, interval: Interval, data: pd.DataFrame, remove_duplicate=True
    ) -> None:
        """
        Stores the stock data in the storage dictionary.
        :symbol: a stock or crypto
        :interval: the interval between each data point, must be at least MIN_1
        :data: a pandas dataframe that has stock data and has a datetime index
        :remove_duplicate: whether to remove data with duplicate timestamps
        """
        # Do not create an entry if there is no data because it will
        # cause the data_range function to fail.
        if data.empty:
            return None

        self.storage_lock.acquire()

        if symbol in self.storage_price:
            # Handles if we already have data
            intervals = self.storage_price[symbol]
            if interval in intervals:
                try:
                    # Handles if we have stock data for the given interval
                    intervals[interval] = self._append(
                        intervals[interval], data, remove_duplicate=remove_duplicate
                    )
                    if self.price_storage_limit:
                        intervals[interval] = intervals[interval][
                            -self.price_storage_size :
                        ]
                except:
                    self.storage_lock.release()
                    raise Exception("Append Failure, case not found!")
            else:
                # Add the data as a new interval
                intervals[interval] = data
        else:
            if self.price_storage_limit:
                data = data[-self.price_storage_size :]
            if len(data) < self.price_storage_size:
                debugger.warning(
                    f"Symbol {symbol}, interval {interval} initialized with only {len(data)} data points"
                )
            # Add the data into storage
            self.storage_price[symbol] = {interval: data}

        cur_len = len(self.storage_price[symbol][interval])
        if self.price_storage_limit and cur_len > self.price_storage_size:
            # If we have more than N data points, remove the oldest data
            self.storage_price[symbol][interval] = self.storage_price[symbol][
                interval
            ].iloc[-self.price_storage_size :]

        self.storage_lock.release()

    def load(
        self,
        symbol: str,
        interval: Interval = None,
        start: dt.datetime = None,
        end: dt.datetime = None,
        slice_data=False,
    ) -> pd.DataFrame:
        """
        Loads the stock data given the symbol and interval. May return only
        a subset of the data if start and end are given.

        If the specified interval does not exist, it will attempt to generate it by
        aggregating data.
        :symbol: a stock or crypto
        :interval: the interval between each data point, must be at least MIN_1
        :start: a datetime object
        """
        self.storage_lock.acquire()

        if symbol not in self.storage_price:
            self.storage_lock.release()
            return None

        if interval is None:
            # If the interval is not given, return the data with the
            # smallest interval that has data in the range.
            intervals = [
                (interval, interval_to_timedelta(interval))
                for interval in self.storage_price[symbol]
            ]
            intervals.sort(key=lambda interval_timedelta: interval_timedelta[1])
            for interval_timedelta in intervals:
                self.storage_lock.release()
                data = self.load(symbol, interval_timedelta[0], start, end)
                self.storage_lock.acquire()
                if data is not None:
                    self.storage_lock.release()
                    return data
            self.storage_lock.release()
            return None

        data = self.storage_price[symbol][interval]

        if start is None and end is None:
            self.storage_lock.release()
            return data

        # If the start and end are not defined, then set them to the
        # beginning and end of the data.
        if start is None:
            start = data.index[0]
        if end is None:
            end = data.index[-1]

        self.storage_lock.release()
        return data.loc[start:end]

    def store_transaction(
        self,
        timestamp: dt.datetime,
        algorithm_name: str,
        symbol: str,
        side: str,
        quantity: int,
        price: float,
    ) -> None:
        self.storage_transaction.append(
            [timestamp, algorithm_name, symbol, side, quantity, price],
            ignore_index=True,
        )

    def reset(self, symbol: str, interval: Interval):
        """
        Resets to an empty dataframe
        """
        self.storage_lock.acquire()
        self.storage_price[symbol][interval] = pd.DataFrame()
        self.storage_lock.release()

    def _append(
        self,
        current_data: pd.DataFrame,
        new_data: pd.DataFrame,
        remove_duplicate: bool = True,
    ) -> pd.DataFrame:
        """
        Appends the data as best it can with gaps in the data for weekends
        and time when no data is collected.
        :current_data: the current data that we have on the stock for
            the interval
        :new_data: data coming from the the broker's API call
        """

        new_df = current_data.append(new_data)
        if remove_duplicate:
            new_df = new_df[~new_df.index.duplicated(keep="last")].sort_index()
        return new_df

    def aggregate(
        self,
        symbol: str,
        base: Interval,
        target: Interval,
        remove_duplicate: bool = True,
    ):
        """
        Aggregates the stock data from the interval specified in 'from' to 'to'.
        """
        self.storage_lock.acquire()
        data = self.storage_price[symbol][base]
        self.storage_price[symbol][target] = self._append(
            self.storage_price[symbol][target],
            aggregate_df(data, target),
            remove_duplicate,
        )
        cur_len = len(self.storage_price[symbol][target])
        if self.price_storage_limit and cur_len > self.price_storage_size:
            self.storage_price[symbol][target] = self.storage_price[symbol][
                target
            ].iloc[-self.price_storage_size :]
        self.storage_lock.release()

    def init_performace_data(self, equity: float, timestamp):
        for interval, days in self.performance_history_intervals:
            self.storage_performance[interval] = pd.DataFrame(
                {"equity": [equity]}, index=[timestamp]
            )

    def add_performance_data(self, equity: float, timestamp):
        """
        Adds the performance data to the storage.

        This function is called every time main() is called in trader.
        It takes the current equity and adds it to each interval.


        :param equity: Current equity of the account.
        """

        # Performance history range up until '3 MONTHS' have the
        # same interval as the polling interval of the trader.
        for interval, days in self.performance_history_intervals[0:3]:
            df = self.storage_performance[interval]
            cutoff = timestamp - dt.timedelta(days=days)
            if df.index[0] < cutoff:
                df = df.loc[df.index >= cutoff]
            df = df.append(pd.DataFrame({"equity": [equity]}, index=[timestamp]))
            self.storage_performance[interval] = df

        # Performance history intervals after '3 MONTHS' are populated
        # only for each day.
        for interval, days in self.performance_history_intervals[3:5]:
            df = self.storage_performance[interval]
            if df.index[-1].date() == timestamp.date():
                df = df.iloc[:-1]
                df = df.append(pd.DataFrame({"equity": [equity]}, index=[timestamp]))
            else:
                df = df.append(pd.DataFrame({"equity": [equity]}, index=[timestamp]))
                cutoff = timestamp - dt.timedelta(days=days)
                if df.index[0] < cutoff:
                    df = df.loc[df.index >= cutoff]
            self.storage_performance[interval] = df

        df = self.storage_performance["ALL"]
        if df.index[-1].date() == timestamp.date():
            df = df.iloc[:-1]
        df = df.append(pd.DataFrame({"equity": [equity]}, index=[timestamp]))
        self.storage_performance["ALL"] = df

        debugger.debug("Performance data added")
        for k in self.storage_performance:
            debugger.debug(f"{k}:\n {self.storage_performance[k]}")



# Builtins
from re import S
import unittest
from unittest.mock import patch

# from harvest import algo
from harvest.trader import PaperTrader
from harvest.api.dummy import DummyStreamer
from harvest.algo import BaseAlgo

from harvest.utils import Interval, gen_data

import logging

prices = [10, 12, 11, 9, 8, 10, 11, 12, 13, 15, 14, 16, 13, 14]


class TestAlgo(unittest.TestCase):

    # Watchlist set in the Algo class locally should take precendance over watchlist in PaperTrader class
    def test_config_watchlist_1(self):
        class Algo1(BaseAlgo):
            def config(self):
                self.watchlist = ["A", "B", "C"]

        class Algo2(BaseAlgo):
            def config(self):
                self.watchlist = ["D", "E", "F"]

        algo1 = Algo1()
        algo2 = Algo2()

        s = DummyStreamer()
        t = PaperTrader(s)
        t.set_algo([algo1, algo2])
        t.set_symbol(["1", "2", "3"])

        t.start()
        s.tick()

        # Getting watchlist should return the watchlist of the Algo class
        list1 = algo1.get_stock_watchlist()
        self.assertListEqual(list1, ["A", "B", "C"])
        list2 = algo2.get_stock_watchlist()
        self.assertListEqual(list2, ["D", "E", "F"])

        # Running methods like get_stock_price_list without a symbol parameter
        # should return the symbol specified in the Algo class
        prices1 = algo1.get_asset_price_list()
        self.assertListEqual(
            prices1, list(t.storage.load("A", Interval.MIN_5)["A"]["close"])
        )

    # PaperTrader class should be able to run algorithms at the individually specified intervals
    def test_config_interval(self):
        class Algo1(BaseAlgo):
            def config(self):
                self.watchlist = ["A"]
                self.interval = "5MIN"
                self.aggregations = ["15MIN", "1DAY"]

        class Algo2(BaseAlgo):
            def config(self):
                self.watchlist = ["B"]
                self.interval = "30MIN"
                self.aggregations = ["1DAY"]

        algo1 = Algo1()
        algo2 = Algo2()

        s = DummyStreamer()
        t = PaperTrader(s)
        t.set_algo([algo1, algo2])

        t.start()
        s.tick()

        self.assertListEqual(
            t.interval["A"]["aggregations"], [Interval.MIN_15, Interval.DAY_1]
        )

    def test_rsi(self):
        """
        Test that RSI values are calculated correctly.
        """
        algo = BaseAlgo()
        algo.add_symbol("DUMMY")
        rsi = algo.rsi(prices=prices)[-1]

        self.assertAlmostEqual(rsi, 59.476113, places=5)

    def test_sma(self):
        """
        Test that SMA values are calculated correctly.
        """
        algo = BaseAlgo()
        algo.add_symbol("DUMMY")
        sma = algo.sma(prices=prices)[-1]

        self.assertAlmostEqual(sma, sum(prices) / len(prices), places=5)

    def test_ema(self):
        """
        Test that EMA values are calculated correctly.
        """
        algo = BaseAlgo()
        algo.add_symbol("DUMMY")
        ema = algo.ema(prices=prices)[-1]

        alpha = 2 / (len(prices) + 1)
        weights = [(1 - alpha) ** t for t in range(len(prices))]
        expected_ema = sum(w * price for w, price in zip(weights, prices[::-1])) / sum(
            weights
        )

        self.assertAlmostEqual(ema, expected_ema, places=5)

    def test_bbands(self):
        """
        Test that bbands returns the correct values based on provided price list.
        """
        algo = BaseAlgo()
        algo.add_symbol("DUMMY")
        upper, middle, lower = algo.bbands(prices=prices)

        mean = sum(prices) / len(prices)
        var = sum((price - mean) ** 2 for price in prices) / (len(prices) - 1)
        std = var ** 0.5
        expected_middle = sum(prices) / len(prices)

        self.assertAlmostEqual(middle[-1], expected_middle, places=5)
        self.assertAlmostEqual(upper[-1], expected_middle + std, places=5)
        self.assertAlmostEqual(lower[-1], expected_middle - std, places=5)

    def test_bbands_trader(self):
        """
        Test that bband values are calculated correctly based on data in PaperTrader's Storage class.
        """
        streamer = DummyStreamer()
        t = PaperTrader(streamer)
        t.set_symbol("DUMMY")
        t.set_algo(BaseAlgo())
        t.start("1MIN")
        streamer.tick()

        upper, middle, lower = t.algo[0].bbands()

        self.assertEqual(True, True)

    def test_get_asset_quantity(self):
        s = DummyStreamer()
        t = PaperTrader(s)
        t.set_symbol("A")
        t.set_algo(BaseAlgo())
        t.start("1MIN")

        # This should buy 5 of A
        t.algo[0].buy("A", 5)
        s.tick()

        q = t.algo[0].get_asset_quantity("A")

        self.assertEqual(q, 5)

    def test_get_asset_cost(self):
        s = DummyStreamer()
        t = PaperTrader(s)
        t.set_symbol("A")
        t.set_algo(BaseAlgo())
        t.start("1MIN")

        t.algo[0].buy("A", 1)
        s.tick()

        cost = s.fetch_price_history("A", Interval.MIN_5).iloc[-1]["A"]["close"]
        get_cost = t.algo[0].get_asset_cost("A")

        self.assertEqual(get_cost, cost)

    def test_get_asset_price(self):
        s = DummyStreamer()
        t = PaperTrader(s)
        t.set_symbol("A")
        t.set_algo(BaseAlgo())
        t.start("1MIN")

        # This should buy 5 of A
        t.algo[0].buy("A", 5)
        s.tick()
        price = s.fetch_latest_stock_price()["A"]["A"]["close"][0]

        get_price = t.algo[0].get_asset_price("A")

        self.assertEqual(get_price, price)

    def test_buy_sell(self):
        s = DummyStreamer()
        t = PaperTrader(s)
        t.set_symbol("A")
        t.set_algo(BaseAlgo())
        t.start("1MIN")

        # This should buy 2 of A
        t.algo[0].buy("A", 2)
        s.tick()

        p = t.stock_positions[0]
        self.assertEqual(p["symbol"], "A")
        self.assertEqual(p["quantity"], 2)

        # This should sell 1 of A
        t.algo[0].sell("A", 1)
        s.tick()

        p = t.stock_positions[0]
        self.assertEqual(p["symbol"], "A")
        self.assertEqual(p["quantity"], 1)

    def test_buy_sell_auto(self):
        s = DummyStreamer()
        t = PaperTrader(s)
        t.set_symbol("A")
        t.set_algo(BaseAlgo())
        t.start("1MIN")

        price = round(t.storage.load("A", Interval.MIN_1)["A"]["close"][-1] * 1.05, 2)
        qty = int(1000000 / price)
        t.algo[0].buy()
        s.tick()

        p = t.stock_positions[0]
        self.assertEqual(p["symbol"], "A")
        self.assertEqual(p["quantity"], qty)

        # This should sell all of A
        t.algo[0].sell()
        s.tick()
        self.assertEqual(0, t.algo[0].get_asset_quantity())

    @patch("harvest.api._base.mark_up")
    def test_buy_sell_option_auto(self, mock_mark_up):
        mock_mark_up.return_value = 10

        streamer = DummyStreamer()
        t = PaperTrader(streamer, debug=True)
        t.set_symbol("X")
        t.set_algo(BaseAlgo())
        t.start("1MIN")
        streamer.tick()

        t.algo[0].buy("X     110101C01000000")
        streamer.tick()

        p = t.option_positions[0]
        self.assertEqual(p["symbol"], "X     110101C01000000")

        t.algo[0].sell_all_options()
        streamer.tick()

        # p = t.stock_positions[0]
        self.assertEqual(0, t.algo[0].get_asset_quantity("X     110101C01000000"))
        # self.assertEqual(p['quantity'], 1)


if __name__ == "__main__":
    unittest.main()



import re
import pandas as pd
import datetime as dt
from typing import Tuple

from sqlalchemy import create_engine, select
from sqlalchemy import Column, Integer, String, DateTime, Float, String
from sqlalchemy.orm import declarative_base, sessionmaker

from harvest.storage import BaseStorage
from harvest.utils import *

"""
This module serves as a storage system for pandas dataframes in with SQL tables.
"""

Base = declarative_base()


class Asset(Base):
    """
    This class defines what is in each row in the Assets table.
    """

    __tablename__ = "asset"
    symbol = Column("symbol", String, primary_key=True)
    interval = Column("interval", String, primary_key=True)
    timestamp = Column("timestamp", DateTime, primary_key=True)
    open_ = Column("open", Float)
    close = Column("close", Float)
    high = Column("high", Float)
    low = Column("low", Float)
    volume = Column("volume", Float)

    def __repr__(self):
        return f"""Asset: {self.timestamp} \t {self.symbol} \t {self.interval} \n 
        open: {self.open_} \t high: {self.high} \t low: {self.low} \t close: {self.close} \t volume: {self.volume}"""


class DBStorage(BaseStorage):
    """
    An extension of the basic storage that saves data in SQL tables.
    """

    def __init__(self, db: str = "sqlite:///data.db"):
        """
        Adds a directory to save data to. Loads any data that is currently in the
        directory.
        """
        engine = create_engine(db)
        Base.metadata.create_all(engine)
        self.Session = sessionmaker(engine)

    def store(
        self,
        symbol: str,
        interval: str,
        data: pd.DataFrame,
        remove_duplicate: bool = True,
    ) -> None:
        """
        Stores the stock data in the storage dictionary in SQL tables.
        :symbol: a stock or crypto
        :interval: the interval between each data point, must be atleast
             1 minute
        :data: a pandas dataframe that has stock data and has a datetime
            index
        """

        if not data.empty:
            data.index = normalize_pandas_dt_index(data)
            data.columns = [column[1] for column in data.columns]
            data.rename(columns={"open": "open_"}, inplace=True)
            data["timestamp"] = data.index
            data["symbol"] = symbol
            data["interval"] = interval
            data = data.to_dict("records")

            with self.Session.begin() as session:
                [session.merge(Asset(**d)) for d in data]
                session.commit()

    def aggregate(
        self, symbol: str, base: str, target: str, remove_duplicate: bool = True
    ):
        """
        Aggregates the stock data from the interval specified in 'from' to 'to'.
        """

        data = self.load(symbol, base)
        agg_data = self._append(
            self.load(symbol, target), aggregate_df(data, target), remove_duplicate
        )
        self.store(symbol, target, agg_data)

    def reset(self, symbol: str, interval: str):
        """
        Resets to an empty dataframe
        """

        with self.Session.begin() as session:
            session.execute(
                Asset.__table__.delete().where(
                    Asset.symbol == symbol and Asset.interval == interval
                )
            )
            session.commit()

    def load(
        self,
        symbol: str,
        interval: str = "",
        start: dt.datetime = None,
        end: dt.datetime = None,
    ) -> pd.DataFrame:
        """
        Loads the stock data given the symbol and interval. May return only
        a subset of the data if start and end are given and there is a gap
        between the last data point and the given end datetime.

        If the specified interval does not exist, it will attempt to generate it by
        aggregating data.
        :symbol: a stock or crypto
        :interval: the interval between each data point, must be at least
             1 minute
        :start: a datetime object
        """

        with self.Session.begin() as session:
            data = session.execute(
                select(
                    Asset.timestamp,
                    Asset.open_,
                    Asset.close,
                    Asset.high,
                    Asset.low,
                    Asset.volume,
                ).where(Asset.symbol == symbol and Asset.interval == interval)
            )
            data = pd.DataFrame(
                data, columns=["timestamp", "open_", "close", "high", "low", "volume"]
            )

            if data.empty:
                return None

            data.set_index("timestamp", inplace=True)
            data.index = data.index.tz_localize(tz="UTC")
            data.rename(columns={"open_": "open"}, inplace=True)
            data.columns = pd.MultiIndex.from_product([[symbol], data.columns])

        # If the start and end are not defined, then set them to the
        # beginning and end of the data.
        if start is None:
            start = data.index[0]
        if end is None:
            end = data.index[-1]

        return data.loc[(data.index >= start) & (data.index <= end)]

    def data_range(self, symbol: str, interval: str) -> Tuple[dt.datetime]:
        return super().data_range(symbol, interval)

    def _append(
        self,
        current_data: pd.DataFrame,
        new_data: pd.DataFrame,
        remove_duplicate: bool = True,
    ) -> pd.DataFrame:
        return super()._append(current_data, new_data, remove_duplicate)



# Builtins
import unittest
import pathlib
import shutil

# Submodule imports
from harvest.trader import BackTester
from harvest.algo import BaseAlgo
from harvest.api.dummy import DummyStreamer
from harvest.utils import *


class TestTester(unittest.TestCase):
    def tear_up_down(func):
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            finally:
                path = pathlib.Path("data")
                shutil.rmtree(path)

        return wrapper

    @tear_up_down
    def test_start_do_nothing(self):
        """Do a quick run-through of the BackTester
        to ensure it can run without crashing.
        """
        s = DummyStreamer()
        t = BackTester(s)
        t.set_symbol("A")
        t.set_algo(BaseAlgo())
        t.start("1MIN", ["5MIN"], period="1DAY")
        self.assertTrue(True)

    @tear_up_down
    def test_check_aggregation(self):
        """ """
        t = BackTester(DummyStreamer())
        t.set_symbol("A")
        t.set_algo(BaseAlgo())
        t.start("1MIN", ["1DAY"], period="1DAY")

        minutes = list(t.storage.load("A", Interval.MIN_1)["A"]["close"])[-200:]
        days_agg = list(
            t.storage.load("A", int(Interval.DAY_1) - 16)["A"]["close"]
        )[-200:]

        self.assertListEqual(minutes, days_agg)

    @tear_up_down
    def test_check_run(self):
        """ """

        class TestAlgo(BaseAlgo):
            def main(self):
                print(self.get_datetime())

        t = BackTester(DummyStreamer())
        t.set_symbol("A")
        t.set_algo(TestAlgo())
        t.start("1MIN", ["1DAY"], period="1DAY")

        self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()



# Builtins
import os
import time
import unittest

from harvest import cli


class TestCLI(unittest.TestCase):
    def test_storage(self):
        for storage in cli.storages.values():
            self.assertTrue(callable(storage))

    def test_streamers(self):
        for streamer in cli.streamers.values():
            self.assertTrue(callable(streamer))

    def test_brokers(self):
        for broker in cli.brokers.values():
            self.assertTrue(callable(broker))

    def test_bad_storage(self):
        try:
            cli._get_storage("I don't exist")
            self.assertTrue(False)
        except ValueError:
            self.assertTrue(True)

    def test_bad_streamer(self):
        try:
            cli._get_streamer("I don't exist")
            self.assertTrue(False)
        except ValueError:
            self.assertTrue(True)

    def test_bad_broekrs(self):
        try:
            cli._get_broker("I don't exist")
            self.assertTrue(False)
        except ValueError:
            self.assertTrue(True)

    def test_start_basic(self):
        crossover = os.path.dirname(os.path.realpath(__file__)) + "/../examples"
        args = cli.parser.parse_args(
            ["start", "-o", "memory", "-s", "dummy", "-b", "paper", "-d", crossover]
        )
        cli.start(args, test=True)
        self.assertTrue(True)

    def test_start_complex(self):
        crossover = os.path.dirname(os.path.realpath(__file__)) + "/../examples"
        args = cli.parser.parse_args(
            ["start", "-o", "pickle", "-s", "yahoo", "-b", "paper", "-d", crossover]
        )
        cli.start(args, test=True)
        self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()



# Builtins
import datetime as dt
from typing import Any, Dict, List, Tuple
import os.path
from pathlib import Path

# External libraries
import pandas as pd
import numpy as np
from tqdm import tqdm
import pytz

# Submodule imports
from harvest.storage import PickleStorage
import harvest.trader.trader as trader
from harvest.api.yahoo import YahooStreamer
from harvest.api.paper import PaperBroker
from harvest.utils import *


class BackTester(trader.PaperTrader):
    """
    This class replaces several key functions to allow backtesting
    on historical data.
    """

    def __init__(self, streamer=None, debug=False, config={}):
        """Initializes the TestTrader."""

        self.streamer = YahooStreamer() if streamer is None else streamer
        self.broker = PaperBroker()

        self.storage = PickleStorage(limit_size=False)  # local cache of historic price
        self._init_attributes()

        self._setup_debugger(debug)

    def start(
        self,
        interval: str = "5MIN",
        aggregations: List[Any] = [],
        source: str = "PICKLE",
        path: str = "./data",
        start=None,
        end=None,
        period=None,
    ):
        """Runs backtesting.

        The interface is very similar to the Trader class, with some additional parameters for specifying
        backtesting configurations.

        :param str? interval: The interval to run the algorithm on. defaults to '5MIN'.
        :param List[str]? aggregations: The aggregations to run. defaults to [].
        :param str? source: The source of backtesting data.
            'FETCH' will pull the latest data using the broker (if specified).
            'CSV' will read data from a locally saved CSV file.
            'PICKLE' will read data from a locally saved pickle file, generated using the Trader class.
            defaults to 'PICKLE'.
        :param str? path: The path to the directory which backtesting data is stored.
            This parameter must be set accordingly if 'source' is set to 'CSV' or 'PICKLE'. defaults to './data'.
        """

        debugger.debug(f"Storing asset data in {path}")
        for a in self.algo:
            a.config()

        self._setup(source, interval, aggregations, path, start, end, period)
        self.broker.setup(self.interval, self.main)
        self.streamer.setup(self.interval, self.main)

        for a in self.algo:
            a.setup()
            a.trader = self

        self.run_backtest()

    def _setup(
        self,
        source: str,
        interval: str,
        aggregations: List,
        path: str,
        start,
        end,
        period,
    ):
        self._setup_params(interval, aggregations)
        self._setup_account()

        self.df = {}

        self.storage.limit_size = False

        start = convert_input_to_datetime(start, self.timezone)
        end = convert_input_to_datetime(end, self.timezone)
        period = convert_input_to_timedelta(period)

        if start is None:
            if end is None:
                start = "MAX" if period is None else "PERIOD"
            else:
                start = end - period
        if end is None:
            if start == "MAX" or start == "PERIOD" or period is None:
                end = "MAX"
            else:
                end = start + period

        if source == "PICKLE":
            self.read_pickle_data()
        elif source == "CSV":
            self.read_csv_data(path)
        else:
            raise Exception(f"Invalid source {source}. Must be 'PICKLE' or 'CSV'")

        common_start = None
        common_end = None
        for s in self.interval:
            for i in [self.interval[s]["interval"]] + self.interval[s]["aggregations"]:
                df = self.storage.load(s, i)
                df = pandas_datetime_to_utc(df, self.timezone)
                if common_start is None or df.index[0] > common_start:
                    common_start = df.index[0]
                if common_end is None or df.index[-1] < common_end:
                    common_end = df.index[-1]

        if start != "MAX" and start != "PERIOD" and start < common_start:
            raise Exception(f"Not enough data is available for a start time of {start}")
        if end != "MAX" and end > common_end:
            raise Exception(
                f"Not enough data is available for an end time of {end}: \nLast datapoint is {common_end}"
            )

        if start == "MAX":
            start = common_start
        elif start == "PERIOD":
            start = common_end - period
        if end == "MAX":
            end = common_end

        self.common_start = start
        self.common_end = end

        print(f"Common start: {start}, common end: {end}")

        for s in self.interval:
            for i in [self.interval[s]["interval"]] + self.interval[s]["aggregations"]:
                df = self.storage.load(s, i).copy()
                df = df.loc[start:end]
                self.storage.reset(s, i)
                self.storage.store(s, i, df)

        conv = {
            Interval.MIN_1: 1,
            Interval.MIN_5: 5,
            Interval.MIN_15: 15,
            Interval.MIN_30: 30,
            Interval.HR_1: 60,
            Interval.DAY_1: 1440,
        }

        # Generate the "simulated aggregation" data
        for sym in self.interval:
            interval = self.interval[sym]["interval"]
            interval_txt = interval_enum_to_string(interval)
            df = self.storage.load(sym, interval)
            df_len = len(df.index)

            debugger.debug(f"Formatting {sym} data...")
            for agg in self.interval[sym]["aggregations"]:
                agg_txt = interval_enum_to_string(agg)
                # tmp_path = f"{path}/{sym}-{interval_txt}+{agg_txt}.pickle"
                tmp_path = f"{path}/{sym}@{int(agg)-16}.pickle"
                file = Path(tmp_path)
                if file.is_file():
                    data = self.storage.open(sym, int(agg) - 16)
                    self.storage.store(sym, int(agg) - 16, data, save_pickle=False)
                    continue
                    # TODO: check if file is updated with latest data
                debugger.debug(
                    f"Formatting aggregation from {interval_txt} to {agg_txt}..."
                )
                points = int(conv[agg] / conv[interval])
                for i in tqdm(range(df_len)):
                    df_tmp = df.iloc[0 : i + 1]
                    df_tmp = df_tmp.iloc[
                        -points:
                    ]  # Only get recent data, since aggregating the entire df will take too long
                    agg_df = aggregate_df(df_tmp, agg)
                    self.storage.store(
                        sym,
                        int(agg) - 16,
                        agg_df.iloc[[-1]],
                        remove_duplicate=False,
                        save_pickle=False,
                    )
        debugger.debug("Formatting complete")
        for sym in self.interval:
            for agg in self.interval[sym]["aggregations"]:
                data = self.storage.load(sym, int(agg) - 16)
                data = pandas_datetime_to_utc(data, self.timezone)
                self.storage.store(
                    sym,
                    int(agg) - 16,
                    data,
                    remove_duplicate=False,
                )

        # # Save the current state of the queue
        # for s in self.watch:
        #     self.load.append_entry(s, self.interval, self.storage.load(s, self.interval))
        #     for i in self.aggregations:
        #         self.load.append_entry(s, '-'+i, self.storage.load(s, '-'+i), False, True)
        #         self.load.append_entry(s, i, self.storage.load(s, i))

        # Move all data to a cached dataframe
        for sym in self.interval:
            self.df[sym] = {}
            inter = self.interval[sym]["interval"]
            interval_txt = interval_enum_to_string(inter)
            df = self.storage.load(sym, inter)
            self.df[sym][inter] = df.copy()

            for agg in self.interval[sym]["aggregations"]:
                # agg_txt = interval_enum_to_string(agg)
                # agg_txt = f"{interval_txt}+{agg_txt}"
                df = self.storage.load(sym, int(agg) - 16)
                self.df[sym][int(agg) - 16] = df.copy()

        # Trim data so start and end dates match between assets and intervals
        # data_start = pytz.utc.localize(dt.datetime(1970, 1, 1))
        # data_end = pytz.utc.localize(dt.datetime.utcnow().replace(microsecond=0, second=0))
        # for i in [self.interval] + self.aggregations:
        #     for s in self.watch:
        #         start = self.df[i][s].index[0]
        #         end = self.df[i][s].index[-1]
        #         if start > data_start:
        #             data_start = start
        #         if end < data_end:
        #             data_end = end

        # for i in [self.interval] + self.aggregations:
        #     for s in self.watch:
        #         self.df[i][s] = self.df[i][s].loc[data_start:data_end]

        self.load_watch = True

    def read_pickle_data(self):
        """Function to read backtesting data from a local file.

        :interval: The interval of the data
        :path: Path to the local data file
        :date_format: The format of the data's timestamps
        """
        for s in self.interval:
            for i in [self.interval[s]["interval"]] + self.interval[s]["aggregations"]:
                df = self.storage.open(s, i).dropna()
                if df.empty or now() - df.index[-1] > dt.timedelta(days=1):
                    df = self.streamer.fetch_price_history(s, i).dropna()
                self.storage.store(s, i, df)

    def read_csv_data(self, path: str, date_format: str = "%Y-%m-%d %H:%M:%S"):
        """Function to read backtesting data from a local CSV file.

        :interval: The interval of the data
        :path: Path to the local data file
        :date_format: The format of the data's timestamps
        """
        for s in self.interval:
            for i in [self.interval[s]["interval"]] + self.interval[s]["aggregations"]:
                i_txt = interval_enum_to_string(i)
                df = self.read_csv(f"{path}/{s}-{i_txt}.csv").dropna()
                if df.empty:
                    df = self.streamer.fetch_price_history(s, i).dropna()
                self.storage.store(s, i, df)

    def read_csv(self, path: str) -> pd.DataFrame:
        """Reads a CSV file and returns a Pandas DataFrame.

        :path: Path to the CSV file.
        """
        if not os.path.isfile(path):
            return pd.DataFrame()
        df = pd.read_csv(path)
        df = df.set_index(["timestamp"])
        if isinstance(df.index[0], str):
            df.index = pd.to_datetime(df.index)
        else:
            df.index = pd.to_datetime(df.index, unit="s")
        df = df[["open", "high", "low", "close", "volume"]].astype(float)
        df = df.sort_index()
        return df

    def run_backtest(self):

        # import cProfile
        # pr = cProfile.Profile()
        # pr.enable()
        # Reset them

        for s in self.interval:
            for i in [self.interval[s]["interval"]] + self.interval[s]["aggregations"]:
                self.storage.reset(s, i)

        self.storage.limit_size = True

        common_start = self.common_start
        common_end = self.common_end

        counter = {}
        for s in self.interval:
            inter = self.interval[s]["interval"]
            start_index = list(self.df[s][inter].index).index(common_start)
            self.interval[s]["start"] = start_index
            counter[s] = 0

        self.timestamp = common_start.to_pydatetime()
        print(f"Starting backtest from {common_start}")

        while self.timestamp <= common_end:
            df_dict = {}
            for sym in self.interval:
                inter = self.interval[sym]["interval"]
                if is_freq(self.timestamp, inter):
                    # If data is not in the cache, skip it
                    if self.timestamp in self.df[sym][inter].index:
                        df_dict[sym] = self.df[sym][inter].loc[self.timestamp]

            update = self._update_order_queue()
            self._update_position_cache(df_dict, new=update, option_update=True)
            for sym in self.interval:
                inter = self.interval[sym]["interval"]
                if is_freq(self.timestamp, inter):

                    # If data is not in the cache, skip it
                    if not self.timestamp in self.df[sym][inter].index:
                        continue
                    df = self.df[sym][inter].loc[[self.timestamp], :]
                    self.storage.store(s, inter, df, save_pickle=False)
                    # Add data to aggregation queue
                    for agg in self.interval[sym]["aggregations"]:
                        df = self.df[s][int(agg) - 16].iloc[
                            [self.interval[sym]["start"] + counter[sym]], :
                        ]
                        self.storage.store(s, agg, df)
                    counter[sym] += 1
            new_algo = []
            for a in self.algo:
                if not is_freq(self.timestamp, a.interval):
                    new_algo.append(a)
                    continue
                try:
                    a.main()
                    new_algo.append(a)
                except Exception as e:
                    debugger.debug(
                        f"Algorithm {a} failed, removing from algorithm list.\nException: {e}"
                    )
            if len(new_algo) == 0:
                debugger.debug(f"No algorithms left, exiting")
                break
            self.algo = new_algo

            self.timestamp += interval_to_timedelta(self.streamer.poll_interval)

        # pr.disable()
        # import pstats
        # st = pstats.Stats(pr)
        # st.sort_stats('cumtime')
        # st.print_stats(0.1)
        # st.dump_stats("stat.txt")

        debugger.debug(self.account)

    def _queue_update(self, new_df: pd.DataFrame, time):
        pass

    def _setup_account(self):
        self.account = {
            "equity": 1000000.0,
            "cash": 1000000.0,
            "buying_power": 1000000.0,
            "multiplier": 1,
        }

    def fetch_position(self, key):
        pass

    def fetch_account(self):
        pass
